# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2013 WandererFan <wandererfan@gmail.com>                *
# *   Copyright (c) 2019 Zheng, Lei (realthunder)<realthunder.dev@gmail.com>*
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the PathArray object.

The copies will be placed along a path like a polyline, spline, or bezier
curve, and along the selected subelements.

To Do
-----
The `'PathSubelements'` property must be changed in type, as it does not need
to be an `App::PropertyLinkSubList`.
A `LinkSubList` is to select multiple subelements (edges) from multiple
objects. However, since we need to select a `'Path Object'` already,
which is a single object, the subelements that we can choose must belong
to this `'Path Object'` only.

Therefore, the correct property that must be used is `App::PropertyLinkSub`.
Then in the property editor we will be unable to select more than one object
thus preventing errors of the subelements not matching the `'PathObject'`.

In fact, both `'PathObject'` and `'PathSubelements'`
could be handled with a single `App::PropertyLinkSub` property,
as this property can be used to select a single object,
or a single object with its subelements.

In the future, we could migrate the properties, or outright break
compatibility with older objects by changing both properties
`'PathObject'` and `'PathSubelements'`.

An alternative to this would be to use a single `App::PropertyLinkSubList`.
This would allow us to build PathArrays on multiple objects and multiple
subelements (edges) of those objects at the same time. However, to do this,
the logic in `execute` would have to be changed to account for multiple
objects. Therefore, the first solution is simpler, that is, using
a single property of type `App::PropertyLinkSub`.
"""
## @package patharray
# \ingroup draftobjects
# \brief Provides the object code for the PathArray object.

import FreeCAD as App
import DraftVecUtils
import lazy_loader.lazy_loader as lz

from draftutils.messages import _err, _log, _wrn
from draftutils.translate import translate


def QT_TRANSLATE_NOOP(ctx, txt):
    return txt


from draftobjects.base import DraftObject
from draftobjects.draftlink import DraftLink

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")

## \addtogroup draftobjects
# @{


class PathArray(DraftLink):
    """The Draft Path Array object.

    The object distributes copies of an object along a path like a polyline,
    spline, or bezier curve.

    Attributes
    ----------
    Align: bool
        It defaults to `False`.
        It sets whether the object will be specially aligned to the path.

    AlignMode: str
        It defaults to `'Original'`.
        Indicates the type of alignment that will be calculated when
        `Align` is `True`.

        `'Original'` mode is the historic `'Align'` for old (v0.18) documents.
        It is not really the Fernat alignment. It uses the normal parameter
        from `getNormal` (or the default) as a constant, it does not calculate
        curve normal.
        `X` is curve tangent, `Y` is normal parameter, `Z` is the cross product
        `X` x `Y`.

        `'Tangent'` mode is similar to `Original`, but includes a pre-rotation
        (in execute) to align the `Base` object's `X` to `TangentVector`,
        then `X` follows curve tangent, normal input parameter
        is the Z component.

        If `ForceVertical` is `True`, the normal parameter from `getNormal`
        is ignored, and `X` is curve tangent, `Z` is `VerticalVector`,
        and `Y` is the cross product `X` x `Z`.

        `'Frenet'` mode orients the copies to a coordinate system
        along the path.
        `X` is tangent to curve, `Y` is curve normal, `Z` is curve binormal.
        If normal cannot be computed, for example, in a straight line,
        the default is used.

    ForceVertical: bool
        It defaults to `False`.
        If it is `True`, and `AlignMode` is `'Original'` or `'Tangent'`,
        it will use the vector in `VerticalVector` as the `Z` axis.

    StartOffset: float
        It defaults to 0.0.
        It is the length from the start of the path to the first copy.

    EndOffset: float
        It defaults to 0.0.
        It is the length at the end of the path that will not be available
        for object placement.

    ReversePath: bool
        It defaults to False.
        This will walk the path in reverse, also reversing object
        orientation. Start and end offsets will count from opposite ends
        of the path, etc.

    SpacingMode: string
        It defaults to `'Fixed count'`.
        Objects can be spaced to divide the available length evenly
        (`'Fixed count'`, this is the original spacing mode from FreeCAD 1.0),
        or to be placed in given distances along the path from each other:
        `'Fixed spacing'` will keep placing objects for as long as there
        is still space available, while `'Fixed count and spacing'`
        will place a given number of objects (provided they fit in available
        space).

    SpacingUnit: length
        It defaults to 20mm.
        When fixed spacing modes are used, this is the spacing distance
        used. If UseSpacingPattern is also enabled, this is the unit length
        of "1.0" in the spacing pattern (so, default pattern of [1.0, 2.0]
        with default SpacingUnit of 20mm means a spacing pattern of
        20mm, 40mm).

    UseSpacingPattern: bool
        Default is False.
        Enables the SpacingPattern for uneven distribution of objects.
        Will have slightly different effect depending on SpacingMode.

    SpacingPattern: float list
        Default is [1.0, 2.0]
        When UseSpacingPattern is True, this list contains the proportions
        of distances between consecutive object pairs. Can be used in any
        spacing mode. In "fixed spacing" modes SpacingPattern is multiplied
        by SpacingUnit. In flexible spacing modes ("fixed count"), spacing
        pattern defines the proportion of distances.
    """

    def __init__(self, obj):
        super().__init__(obj, "PathArray")

    def attach(self, obj):
        """Set up the properties when the object is attached.

        Note: we don't exactly know why the properties are added
        in the `attach` method. They should probably be added in the `__init__`
        method. Maybe this is related to the link behavior of this class.

        For PathLinkArray, DraftLink.attach creates the link to the Base.

        Realthunder: before the big merge, there was only the attach() method
        in the view object proxy, not the object proxy.
        I added that to allow the proxy to override the C++ view provider
        type. The view provider type is normally determined by object's
        C++ API getViewProviderName(), and cannot be overridden by the proxy.
        I introduced the attach() method in proxy to allow the core
        to attach the proxy before creating the C++ view provider.
        """
        self.set_properties(obj)
        super().attach(obj)

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        if not obj:
            return

        if hasattr(obj, "PropertiesList"):
            properties = obj.PropertiesList
        else:
            properties = []

        self.set_general_properties(obj, properties)
        self.set_spacing_properties(obj, properties)
        self.set_align_properties(obj, properties)

    def set_general_properties(self, obj, properties):
        """Set general properties only if they don't exist."""
        if "Base" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "The base object that will be duplicated")
            obj.addProperty("App::PropertyLinkGlobal", "Base", "Objects", _tip, locked=True)
            obj.Base = None

        if "PathObject" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The object along which the copies will be distributed. It must contain 'Edges'.",
            )
            obj.addProperty("App::PropertyLinkGlobal", "PathObject", "Objects", _tip, locked=True)
            obj.PathObject = None

        # TODO: the 'PathSubelements' property must be changed,
        # as it does not need to be an 'App::PropertyLinkSubList'.
        #
        # In fact, both 'PathObject' and 'PathSubelements'
        # could be handled with a single 'App::PropertyLinkSub' property,
        # as this property can be used to select a single object,
        # or a single object with its subelements.
        if "PathSubelements" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "List of connected edges in the 'Path Object'.\nIf these are present, the copies will be created along these subelements only.\nLeave this property empty to create copies along the entire 'Path Object'.",
            )
            obj.addProperty(
                "App::PropertyLinkSubListGlobal", "PathSubelements", "Objects", _tip, locked=True
            )
            obj.PathSubelements = []

        if "Fuse" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Specifies if the copies "
                "should be fused together "
                "if they touch each other (slower)",
            )
            obj.addProperty("App::PropertyBool", "Fuse", "Objects", _tip, locked=True)
            obj.Fuse = False

        if self.use_link and "ExpandArray" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Show the individual array elements (only for Link arrays)"
            )
            obj.addProperty("App::PropertyBool", "ExpandArray", "Objects", _tip, locked=True)
            obj.ExpandArray = False
            obj.setPropertyStatus("Shape", "Transient")

        if not self.use_link:
            if "PlacementList" not in properties:
                _tip = QT_TRANSLATE_NOOP("App::Property", "The placement for each array element")
                obj.addProperty(
                    "App::PropertyPlacementList", "PlacementList", "Objects", _tip, locked=True
                )
                obj.PlacementList = []

    def set_align_properties(self, obj, properties):
        """Set general properties only if they don't exist."""
        if "ExtraTranslation" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Additional translation that will be applied to each copy.\nThis is useful to adjust for the difference between shape centre and shape reference point.",
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "ExtraTranslation", "Alignment", _tip, locked=True
            )
            obj.ExtraTranslation = App.Vector(0, 0, 0)

        if "TangentVector" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Alignment vector for 'Tangent' mode")
            obj.addProperty("App::PropertyVector", "TangentVector", "Alignment", _tip, locked=True)
            obj.TangentVector = App.Vector(1, 0, 0)

        if "ForceVertical" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Force use of 'Vertical Vector' as local Z-direction when using 'Original' or 'Tangent' alignment mode",
            )
            obj.addProperty("App::PropertyBool", "ForceVertical", "Alignment", _tip, locked=True)
            obj.ForceVertical = False

        if "VerticalVector" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Direction of the local Z axis when 'Force Vertical' is true"
            )
            obj.addProperty("App::PropertyVector", "VerticalVector", "Alignment", _tip, locked=True)
            obj.VerticalVector = App.Vector(0, 0, 1)

        if "AlignMode" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Method to orient the copies along the path.\n- Original: X is curve tangent, Y is normal, and Z is the cross product.\n- Frenet: aligns the object following the local coordinate system along the path.\n- Tangent: similar to 'Original' but the local X axis is pre-aligned to 'Tangent Vector'.\n\nTo get better results with 'Original' or 'Tangent' you may have to set 'Force Vertical' to true.",
            )
            obj.addProperty("App::PropertyEnumeration", "AlignMode", "Alignment", _tip, locked=True)
            obj.AlignMode = ["Original", "Frenet", "Tangent"]
            obj.AlignMode = "Original"

        if "ReversePath" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Walk the path backwards.")
            obj.addProperty("App::PropertyBool", "ReversePath", "Alignment", _tip, locked=True)
            obj.ReversePath = False

        # The Align property must be attached after other align properties
        # so that onChanged works properly
        if "Align" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Orient the copies along the path depending on the 'Align Mode'.\nOtherwise the copies will have the same orientation as the original Base object.",
            )
            obj.addProperty("App::PropertyBool", "Align", "Alignment", _tip, locked=True)
            obj.Align = False

    def set_spacing_properties(self, obj, properties):

        if "Count" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Number of copies to create")
            obj.addProperty("App::PropertyInteger", "Count", "Spacing", _tip, locked=True)
            obj.Count = 4

        if "SpacingMode" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "How copies are spaced.\n"
                + " - Fixed count: available path length (minus start and end offsets) is evenly divided into n.\n"
                + ' - Fixed spacing: start at "Start offset" and place new copies after traveling a fixed distance along the path.\n'
                + ' - Fixed count and spacing: same as "Fixed spacing", but also stop at given number of copies.',
            )
            obj.addProperty("App::PropertyEnumeration", "SpacingMode", "Spacing", _tip, locked=True)
            obj.SpacingMode = ["Fixed count", "Fixed spacing", "Fixed count and spacing"]
            obj.SpacingMode = "Fixed count"

        if "SpacingUnit" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Base fixed distance between elements.")
            obj.addProperty("App::PropertyLength", "SpacingUnit", "Spacing", _tip, locked=True)
            obj.SpacingUnit = 20.0
            obj.setPropertyStatus("SpacingUnit", "Hidden")

        if "UseSpacingPattern" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Use repeating spacing patterns instead of uniform spacing."
            )
            obj.addProperty("App::PropertyBool", "UseSpacingPattern", "Spacing", _tip, locked=True)
            obj.UseSpacingPattern = False

        if "SpacingPattern" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Spacing is multiplied by a corresponding number in this sequence."
            )
            obj.addProperty(
                "App::PropertyFloatList", "SpacingPattern", "Spacing", _tip, locked=True
            )
            obj.SpacingPattern = [1, 2]
            obj.setPropertyStatus("SpacingPattern", "Hidden")

        if "StartOffset" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Length from the start of the path to the first copy."
            )
            obj.addProperty("App::PropertyLength", "StartOffset", "Spacing", _tip, locked=True)
            obj.StartOffset = 0.0

        if "EndOffset" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Length from the end of the path to the last copy."
            )
            obj.addProperty("App::PropertyLength", "EndOffset", "Spacing", _tip, locked=True)
            obj.EndOffset = 0.0

    def linkSetup(self, obj):
        """Set up the object as a link object."""
        super().linkSetup(obj)
        obj.configLinkProperty(ElementCount="Count")

    def execute(self, obj):
        """Execute when the object is created or recomputed."""
        if self.props_changed_placement_only(obj) or not obj.Base or not obj.PathObject:
            self.props_changed_clear()
            return

        # placement of entire PathArray object
        array_placement = obj.Placement

        w = self.get_wires(obj.PathObject, obj.PathSubelements)
        if not w:
            _err(obj.PathObject.Label + translate("draft", ", path object does not have 'Edges'."))
            return

        base_rotation = obj.Base.Shape.Placement.Rotation
        final_rotation = base_rotation

        if obj.Align and obj.AlignMode == "Tangent" and hasattr(obj, "TangentVector"):
            Xaxis = App.Vector(1.0, 0.0, 0.0)  # default TangentVector

            if not DraftVecUtils.equals(Xaxis, obj.TangentVector):
                # make rotation from TangentVector to X
                pre_rotation = App.Rotation(obj.TangentVector, Xaxis)
                final_rotation = base_rotation.multiply(pre_rotation)

        copy_placements = placements_on_path(
            final_rotation,
            w,
            obj.Count,
            obj.ExtraTranslation,
            obj.Align,
            obj.AlignMode,
            obj.ForceVertical,
            obj.VerticalVector,
            obj.StartOffset.Value,
            obj.EndOffset.Value,
            obj.ReversePath,
            obj.SpacingMode,
            obj.SpacingUnit.Value,
            obj.UseSpacingPattern,
            obj.SpacingPattern,
        )

        self.buildShape(obj, array_placement, copy_placements)
        self.props_changed_clear()
        return not self.use_link

    def get_wires(self, path_object, subelements):
        """Get wires from the path object."""
        if subelements:
            w = self.get_wire_from_subelements(subelements)
        elif hasattr(path_object.Shape, "Wires") and path_object.Shape.Wires:
            w = path_object.Shape.Wires[0]
        elif path_object.Shape.Edges:
            w = Part.Wire(path_object.Shape.Edges)
        else:
            w = None
        return w

    def get_wire_from_subelements(self, subelements):
        """Make a wire from the path object subelements."""
        sl = []
        for sub in subelements:
            edgeNames = sub[1]
            for n in edgeNames:
                e = sub[0].Shape.getElement(n)
                sl.append(e)
        return Part.Wire(Part.__sortEdges__(sl))

    def onChanged(self, obj, prop):
        """Execute when a property is changed."""
        super().onChanged(obj, prop)
        self.show_and_hide(obj, prop)

    def show_and_hide(self, obj, prop):
        """Show and hide the properties depending on the touched property.

        Note that when the array is created, some properties will change
        more than once in a seemingly random order.
        """
        # The minus sign removes the Hidden property (show).

        if prop == "SpacingMode":

            # Check if all referenced properties are available:
            for pr in ("SpacingMode", "SpacingUnit", "UseSpacingPattern", "SpacingPattern"):
                if not hasattr(obj, pr):
                    return

            if obj.SpacingMode == "Fixed spacing":
                obj.setPropertyStatus("Count", "Hidden")
                obj.setPropertyStatus("SpacingUnit", "-Hidden")

            elif obj.SpacingMode == "Fixed count":
                obj.setPropertyStatus("Count", "-Hidden")
                obj.setPropertyStatus("SpacingUnit", "Hidden")

            elif obj.SpacingMode == "Fixed count and spacing":
                obj.setPropertyStatus("Count", "-Hidden")
                obj.setPropertyStatus("SpacingUnit", "-Hidden")

        if prop == "UseSpacingPattern":

            # Check if referenced property is available:
            if not hasattr(obj, "SpacingPattern"):
                return

            if obj.UseSpacingPattern:
                obj.setPropertyStatus("SpacingPattern", "-Hidden")
            else:
                obj.setPropertyStatus("SpacingPattern", "Hidden")

        if prop in ("Align", "AlignMode"):

            # Check if all referenced properties are available:
            for pr in ("Align", "AlignMode", "ForceVertical", "VerticalVector", "TangentVector"):
                if not hasattr(obj, pr):
                    return

            if obj.Align:
                obj.setPropertyStatus("AlignMode", "-Hidden")

                if obj.AlignMode == "Frenet":
                    for pr in ("ForceVertical", "VerticalVector"):
                        obj.setPropertyStatus(pr, "Hidden")
                else:
                    for pr in ("ForceVertical", "VerticalVector"):
                        obj.setPropertyStatus(pr, "-Hidden")

                if obj.AlignMode == "Tangent":
                    obj.setPropertyStatus("TangentVector", "-Hidden")
                else:
                    obj.setPropertyStatus("TangentVector", "Hidden")

            else:
                for pr in ("AlignMode", "ForceVertical", "VerticalVector", "TangentVector"):
                    obj.setPropertyStatus(pr, "Hidden")

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        # ReversePath was added together with several Spacing properties in v1.1,
        # and PlacementList property was added for non-link arrays in v1.1,
        # obj should be OK if both are present:
        if hasattr(obj, "ReversePath") and hasattr(obj, "PlacementList"):
            return

        if hasattr(obj, "PathObj"):
            _log("v0.19, " + obj.Name + ", migrated 'PathObj' property to 'PathObject'")
        if hasattr(obj, "PathSubs"):
            _log("v0.19, " + obj.Name + ", migrated 'PathSubs' property to 'PathSubelements'")
        if hasattr(obj, "Xlate"):
            _log("v0.19, " + obj.Name + ", migrated 'Xlate' property to 'ExtraTranslation'")
        if not hasattr(obj, "Fuse"):
            _log("v1.0, " + obj.Name + ", added 'Fuse' property")
        if obj.getGroupOfProperty("Count") != "Spacing":
            _log("v1.1, " + obj.Name + ", moved 'Count' property to 'Spacing' subsection")
        if not hasattr(obj, "ReversePath"):
            _log(
                "v1.1, "
                + obj.Name
                + ", "
                + "added 'ReversePath', 'SpacingMode', 'SpacingUnit', 'UseSpacingPattern' "
                + "and 'SpacingPattern' properties"
            )
        if not hasattr(obj, "PlacementList"):
            _log("v1.1, " + obj.Name + ", added hidden property 'PlacementList'")

        self.set_properties(obj)
        obj.setGroupOfProperty("Count", "Spacing")
        if hasattr(obj, "PathObj"):
            obj.PathObject = obj.PathObj
            obj.removeProperty("PathObj")
        if hasattr(obj, "PathSubs"):
            obj.PathSubelements = obj.PathSubs
            obj.removeProperty("PathSubs")
        if hasattr(obj, "Xlate"):
            obj.ExtraTranslation = obj.Xlate
            obj.removeProperty("Xlate")
        self.execute(obj)  # Required to update PlacementList.


# Alias for compatibility with v0.18 and earlier
_PathArray = PathArray


def placements_on_path(
    shapeRotation,
    pathwire,
    count,
    xlate,
    align,
    mode="Original",
    forceNormal=False,
    normalOverride=None,
    startOffset=0.0,
    endOffset=0.0,
    reversePath=False,
    spacingMode="Fixed count",
    spacingUnit=20.0,
    useSpacingPattern=False,
    spacingPattern=[1, 1, 1, 1],
):
    """Calculate the placements of a shape along a given path.

    Copies will be distributed according to spacing mode - evenly or in fixed offsets.
    """

    if mode == "Frenet":
        forceNormal = False

    if forceNormal and normalOverride:
        normal = normalOverride
    else:
        normal = DraftGeomUtils.get_shape_normal(pathwire)
        if normal is None:
            normal = App.Vector(0, 0, 1)

    path = Part.__sortEdges__(pathwire.Edges)

    # if ReversePath is on, walk the path backwards:
    if reversePath:
        path = path[::-1]

    # find cumulative edge end distance
    totalDist = 0
    ends = []
    for e in path:
        totalDist += e.Length
        ends.append(totalDist)

    # if align is True the length of the path cannot be zero:
    minLength = 1e-6 if align else -1e-12

    if startOffset > (totalDist - minLength):
        if startOffset != 0:
            _wrn(translate("draft", "Start Offset too large for path length. Using 0 instead."))
        startOffset = 0

    if endOffset > (totalDist - startOffset - minLength):
        if endOffset != 0:
            _wrn(
                translate(
                    "draft",
                    "End Offset too large for path length minus Start Offset. Using 0 instead.",
                )
            )
        endOffset = 0

    totalDist = totalDist - startOffset - endOffset

    useFlexibleSpacing = spacingMode == "Fixed count"
    useFixedSpacing = spacingMode in ("Fixed spacing", "Fixed count and spacing")

    stopAfterCount = spacingMode in ("Fixed count", "Fixed count and spacing")
    stopAfterDistance = spacingMode in ("Fixed spacing", "Fixed count and spacing")

    spacingUnit = max(spacingUnit, 0)
    # protect from infinite loop when step = 0
    if spacingUnit == 0:
        _wrn(translate("draft", "Spacing unit of 0 is not allowed, using default"))
        spacingUnit = totalDist

    # negative spacing steps are not defined
    spacingPattern = [abs(w) for w in spacingPattern]

    # protect from infinite loop when pattern weights are all zeros
    if sum(spacingPattern) == 0:
        spacingPattern = [spacingUnit]

    isClosedPath = DraftGeomUtils.isReallyClosed(pathwire) and not (startOffset or endOffset)

    count = max(count, 1)

    if useFlexibleSpacing:
        # Spaces between objects will stretch to fill available length

        segCount = count if isClosedPath else count - 1
        segCount = max(segCount, 1)

        if useSpacingPattern:
            # Available length will be non-uniformly divided in proportions from SpacingPattern:
            fullSpacingPattern = [spacingPattern[i % len(spacingPattern)] for i in range(segCount)]
            sumWeights = sum(fullSpacingPattern)
            distPerWeightUnit = totalDist / sumWeights
            steps = [distPerWeightUnit * weight for weight in fullSpacingPattern]

        else:
            # Available length will be evenly divided (the original spacing method):
            steps = [totalDist / segCount]

    if useFixedSpacing:
        # Objects will be placed in specified intervals

        if useSpacingPattern:
            # Intervals will be fixed, but follow a repeating pattern:
            steps = [spacingUnit * mult for mult in spacingPattern]
        else:
            # Each interval will be the same:
            steps = [spacingUnit]

    remains = 0
    travel = startOffset
    endTravel = startOffset + totalDist
    placements = []

    i = 0
    while True:
        # which edge in path should contain this shape?
        for j in range(len(ends)):
            if travel <= ends[j]:
                iend = j
                remains = ends[iend] - travel
                offset = path[iend].Length - remains if not reversePath else remains
                break
        else:
            # avoids problems with float math travel > ends[-1]
            iend = len(ends) - 1
            offset = path[iend].Length if not reversePath else 0

        # place shape at proper spot on proper edge
        pt = path[iend].valueAt(get_parameter_from_v0(path[iend], offset))
        place = calculate_placement(
            shapeRotation,
            path[iend],
            offset,
            pt,
            xlate,
            align,
            normal,
            mode,
            forceNormal,
            reversePath,
        )
        placements.append(place)
        travel += steps[i % len(steps)]
        i = i + 1

        # End conditions:
        if stopAfterDistance and travel > endTravel:
            break
        if stopAfterCount and i >= count:
            break

        # Failsafe:
        if i > 10_000:
            _wrn(translate("draft", "Operation would generate too many objects. Aborting"))
            return placements[0:1]

    return placements


calculatePlacementsOnPath = placements_on_path


def calculate_placement(
    globalRotation,
    edge,
    offset,
    RefPt,
    xlate,
    align,
    normal=App.Vector(0.0, 0.0, 1.0),
    mode="Original",
    overrideNormal=False,
    reversePath=False,
):
    """Orient shape in the local coordinate system at parameter offset.

    http://en.wikipedia.org/wiki/Euler_angles (previous version)
    http://en.wikipedia.org/wiki/Quaternions
    """
    # Default Placement:
    placement = App.Placement()
    placement.Rotation = globalRotation.inverted() if reversePath else globalRotation
    placement.Base = RefPt + placement.Rotation.multVec(xlate)

    if not align:
        return placement

    tol = 1e-6  # App.Rotation() tolerance is 1e-7. Shorter vectors are ignored.
    nullv = App.Vector()

    t = edge.tangentAt(get_parameter_from_v0(edge, offset))

    if t.isEqual(nullv, tol):
        _wrn(translate("draft", "Length of tangent vector is 0. Copy not aligned."))
        return placement

    if reversePath:
        t.multiply(-1)

    # If the length of the normal is zero or if it is parallel to the tangent,
    # we make the vectors equal (n = t). The App.Rotation() algorithm will
    # then replace the normal with a default axis.
    # For the vector with the lowest App.Rotation() priority we use a null
    # vector. Calculating this binormal would not make sense in the mentioned
    # cases. And in all other cases calculating it is not necessary as
    # App.Rotation() will ignore it.

    if mode in ("Original", "Tangent"):
        n = normal
        if n.isEqual(nullv, tol):
            _wrn(translate("draft", "Length of normal vector is 0. Using a default axis instead."))
            n = t
        else:
            n_nor = n.normalize()
            t_nor = t.normalize()
            if n_nor.isEqual(t_nor, tol) or n_nor.isEqual(t_nor.negative(), tol):
                _wrn(
                    translate(
                        "draft",
                        "Tangent and normal vectors are parallel. Normal replaced by a default axis.",
                    )
                )
                n = t

        if overrideNormal:
            onPathRotation = App.Rotation(t, nullv, n, "XZY")  # priority = "XZY"
        else:
            onPathRotation = App.Rotation(t, n, nullv, "XYZ")  # priority = "XYZ"

    elif mode == "Frenet":
        try:
            n = edge.normalAt(get_parameter_from_v0(edge, offset))
        except App.Base.FreeCADError:  # no/infinite normals here
            _wrn(
                translate(
                    "draft", "Cannot calculate normal vector. Using the default normal instead."
                )
            )
            n = normal

        if n.isEqual(nullv, tol):
            _wrn(translate("draft", "Length of normal vector is 0. Using a default axis instead."))
            n = t
        else:
            n_nor = n.normalize()
            t_nor = t.normalize()
            if n_nor.isEqual(t_nor, tol) or n_nor.isEqual(t_nor.negative(), tol):
                _wrn(
                    translate(
                        "draft",
                        "Tangent and normal vectors are parallel. Normal replaced by a default axis.",
                    )
                )
                n = t

        onPathRotation = App.Rotation(t, n, nullv, "XYZ")  # priority = "XYZ"

    else:
        _err(translate("draft", "AlignMode {} is not implemented").format(mode))
        return placement

    placement.Rotation = onPathRotation.multiply(globalRotation)
    placement.Base = RefPt + placement.Rotation.multVec(xlate)

    return placement


calculatePlacement = calculate_placement


def get_parameter_from_v0(edge, offset):
    """Return parameter at distance offset from edge.Vertexes[0].

    sb method in Part.TopoShapeEdge???
    """
    lpt = edge.valueAt(edge.getParameterByLength(0))
    vpt = edge.Vertexes[0].Point

    if not DraftVecUtils.equals(vpt, lpt):
        # this edge is flipped
        length = edge.Length - offset
    else:
        # this edge is right way around
        length = offset

    return edge.getParameterByLength(length)


getParameterFromV0 = get_parameter_from_v0

## @}



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import sys
import FreeCAD
from FreeCAD import Vector

import ObjectsFem
import Part
import Sketcher

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Deformation (nonlinear elasticity) - Elmer",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["displacement", "spring"],
        "solvers": ["elmer"],
        "material": "solid",
        "equations": ["deformation"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.equation_deformation_spring_elmer import setup
setup()

Deformation equation - Elmer solver

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects

    # sketch defining the spring form
    body = doc.addObject("PartDesign::Body", "Body")
    SketchPath = body.newObject("Sketcher::SketchObject", "Spring_Path")
    SketchPath.AttachmentSupport = (doc.getObject("XY_Plane"), [""])
    SketchPath.MapMode = "FlatFace"
    SketchPath.addGeometry(
        Part.LineSegment(Vector(-20.0, 30.0, 0.0), Vector(-20.0, 0.0, 0.0)), False
    )
    SketchPath.addConstraint(Sketcher.Constraint("PointOnObject", 0, 2, -1))
    SketchPath.addConstraint(Sketcher.Constraint("Vertical", 0))
    SketchPath.addGeometry(
        Part.ArcOfCircle(
            Part.Circle(Vector(0.0, 0.0, 0.0), Vector(0, 0, 1), 20.0),
            3.141593,
            6.283185,
        ),
        False,
    )
    SketchPath.addConstraint(Sketcher.Constraint("Tangent", 0, 2, 1, 1))
    SketchPath.addConstraint(Sketcher.Constraint("PointOnObject", 1, 2, -1))
    SketchPath.addGeometry(Part.LineSegment(Vector(20.0, 0.0, 0.0), Vector(20.0, 30.0, 0.0)), False)
    SketchPath.addConstraint(Sketcher.Constraint("Tangent", 1, 2, 2, 1))
    SketchPath.addConstraint(Sketcher.Constraint("Equal", 2, 0))
    SketchPath.ViewObject.Visibility = False

    # sketch defining the spring cross section
    SketchCircle = body.newObject("Sketcher::SketchObject", "Spring_Circle")
    SketchCircle.AttachmentSupport = (doc.getObject("XZ_Plane"), [""])
    SketchCircle.MapMode = "FlatFace"
    SketchCircle.addGeometry(Part.Circle(Vector(-20.0, 0.0, 0.0), Vector(0, 0, 1), 7.5), False)
    SketchCircle.addConstraint(Sketcher.Constraint("PointOnObject", 0, 3, -1))
    SketchCircle.ViewObject.Visibility = False

    # the spring object
    SpringObject = body.newObject("PartDesign::AdditivePipe", "Spring")
    SpringObject.Profile = SketchCircle
    SpringObject.Spine = SketchPath

    # set view
    doc.recompute()
    if FreeCAD.GuiUp:
        SpringObject.ViewObject.Document.activeView().viewTop()
        SpringObject.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")
    if FreeCAD.GuiUp:
        import FemGui

        FemGui.setActiveAnalysis(analysis)

    # solver
    if solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        solver_obj.OutputIntervals = [4]
        solver_obj.SimulationType = "Transient"
        solver_obj.TimestepIntervals = [21]
        solver_obj.TimestepSizes = [0.05]
        equation_deformation = ObjectsFem.makeEquationDeformation(doc, solver_obj)
        equation_deformation.setExpression("LinearTolerance", "1e-7")
        equation_deformation.LinearIterativeMethod = "Idrs"
        equation_deformation.LinearPreconditioning = "ILU1"
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
        return doc
    analysis.addObject(solver_obj)

    # material iron
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Iron")
    mat = material_obj.Material
    mat["Name"] = "Iron Generic"
    mat["YoungsModulus"] = "211 GPa"
    mat["PoissonRatio"] = "0.29"
    mat["Density"] = "7874 kg/m^3"
    material_obj.Material = mat
    material_obj.References = [(body, "Solid1")]
    analysis.addObject(material_obj)

    # constraints displacement
    DisplaceLeft = doc.addObject("Fem::ConstraintDisplacement", "DisplacementLeft")
    DisplaceLeft.xFree = False
    DisplaceLeft.hasXFormula = True
    DisplaceLeft.xDisplacementFormula = 'Variable "time"; Real MATC "0.006*tx"'
    DisplaceLeft.yFree = False
    DisplaceLeft.yDisplacement = 0
    DisplaceLeft.zFree = False
    DisplaceLeft.zDisplacement = 0
    DisplaceLeft.References = [(SpringObject, "Face1")]
    analysis.addObject(DisplaceLeft)

    DisplaceRight = doc.addObject("Fem::ConstraintDisplacement", "DisplacementRight")
    DisplaceRight.xFree = False
    DisplaceRight.hasXFormula = True
    DisplaceRight.xDisplacementFormula = 'Variable "time"; Real MATC "-0.006*tx"'
    DisplaceRight.yFree = False
    DisplaceRight.yDisplacement = 0
    DisplaceRight.zFree = False
    DisplaceRight.zDisplacement = 0
    DisplaceRight.References = [(SpringObject, "Face5")]
    analysis.addObject(DisplaceRight)

    # constraints spring
    StiffnessLeft = doc.addObject("Fem::ConstraintSpring", "StiffnessLeft")
    StiffnessLeft.TangentialStiffness = "50 N/m"
    StiffnessLeft.ElmerStiffness = "Tangential Stiffness"
    StiffnessLeft.References = [(SpringObject, "Face1")]
    analysis.addObject(StiffnessLeft)

    StiffnessRight = doc.addObject("Fem::ConstraintSpring", "StiffnessRight")
    StiffnessRight.TangentialStiffness = "50 N/m"
    StiffnessRight.ElmerStiffness = "Tangential Stiffness"
    StiffnessRight.References = [(SpringObject, "Face5")]
    analysis.addObject(StiffnessRight)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = body
    femmesh_obj.CharacteristicLengthMax = "1.25 mm"
    femmesh_obj.ElementOrder = "1st"
    femmesh_obj.ViewObject.Visibility = False

    # generate the mesh
    generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2023 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Contains the parser class for extracting metadata from a FreeCAD macro"""

# pylint: disable=too-few-public-methods

import io
import re
from typing import Any, Tuple

import addonmanager_freecad_interface as fci

try:
    from PySide import QtCore
except ImportError:
    QtCore = None

try:
    from addonmanager_licenses import get_license_manager
except ImportError:
    get_license_manager = None


class DummyThread:
    @classmethod
    def isInterruptionRequested(cls):
        return False


class MacroParser:
    """Extracts metadata information from a FreeCAD macro"""

    MAX_LINES_TO_SEARCH = 500  # To speed up parsing: some files are VERY large

    def __init__(self, name: str, code: str = ""):
        """Create a parser for the macro named "name". Note that the name is only
        used as the context for error messages, it is not otherwise important."""
        self.name = name
        self.parse_results = {
            "comment": "",
            "url": "",
            "wiki": "",
            "version": "",
            "other_files": [""],
            "author": "",
            "date": "",
            "license": "",
            "icon": "",
            "xpm": "",
        }
        self.remaining_item_map = {}
        self.current_thread = DummyThread() if QtCore is None else QtCore.QThread.currentThread()
        if code:
            self.fill_details_from_code(code)

    def _reset_map(self):
        """This map tracks which items we've already read. If the same parser is used
        twice, it has to be reset."""
        self.remaining_item_map = {
            "__comment__": "comment",
            "__web__": "url",
            "__wiki__": "wiki",
            "__version__": "version",
            "__files__": "other_files",
            "__author__": "author",
            "__date__": "date",
            "__license__": "license",
            "__licence__": "license",  # accept either spelling
            "__icon__": "icon",
            "__xpm__": "xpm",
        }

    def fill_details_from_code(self, code: str) -> None:
        """Reads in the macro code from the given string and parses it for its
        metadata."""

        self._reset_map()
        line_counter = 0
        content_lines = io.StringIO(code)
        while content_lines and line_counter < self.MAX_LINES_TO_SEARCH:
            line = content_lines.readline()
            if not line:
                break
            if self.current_thread.isInterruptionRequested():
                return
            line_counter += 1
            if not line.startswith("__"):
                # Speed things up a bit... this comparison is very cheap
                continue
            try:
                self._process_line(line, content_lines)
            except SyntaxError as e:
                err_string = f"Syntax error when parsing macro {self.name}:\n{str(e)}"
                fci.Console.PrintWarning(err_string)

    def _process_line(self, line: str, content_lines: io.StringIO):
        """Given a single line of the macro file, see if it matches one of our items,
        and if so, extract the data."""

        lowercase_line = line.lower()
        for key in self.remaining_item_map:
            if lowercase_line.startswith(key):
                self._process_key(key, line, content_lines)
                break

    def _process_key(self, key: str, line: str, content_lines: io.StringIO):
        """Given a line that starts with a known key, extract the data for that key,
        possibly reading in additional lines (if it contains a line continuation
         character or is a triple-quoted string)."""

        line = self._handle_backslash_continuation(line, content_lines)
        line, was_triple_quoted = self._handle_triple_quoted_string(line, content_lines)

        _, _, line = line.partition("=")
        if not was_triple_quoted:
            line, _, _ = line.partition("#")
        self._detect_illegal_content(line)
        final_content_line = line.strip()

        stripped_of_quotes = self._strip_quotes(final_content_line)
        if stripped_of_quotes is not None:
            self._standard_extraction(self.remaining_item_map[key], stripped_of_quotes)
            self.remaining_item_map.pop(key)
        else:
            self._apply_special_handling(key, line)

    @staticmethod
    def _handle_backslash_continuation(line, content_lines) -> str:
        while line.strip().endswith("\\"):
            line = line.strip()[:-1]
            concat_line = content_lines.readline()
            line += concat_line.strip()
        return line

    @staticmethod
    def _handle_triple_quoted_string(line, content_lines) -> Tuple[str, bool]:
        result = line
        was_triple_quoted = False
        if '"""' in result:
            was_triple_quoted = True
            while True:
                new_line = content_lines.readline()
                if not new_line:
                    raise SyntaxError("Syntax error while reading macro")
                if '"""' in new_line:
                    last_line, _, _ = new_line.partition('"""')
                    result += last_line + '"""'
                    break
                result += new_line
        return result, was_triple_quoted

    @staticmethod
    def _strip_quotes(line) -> str:
        line = line.strip()
        stripped_of_quotes = None
        if line.startswith('"""') and line.endswith('"""'):
            stripped_of_quotes = line[3:-3]
        elif (line[0] == '"' and line[-1] == '"') or (line[0] == "'" and line[-1] == "'"):
            stripped_of_quotes = line[1:-1]
        elif (line.startswith('u"') and line[-1] == '"') or (
            line.startswith("u'") and line[-1] == "'"
        ):
            stripped_of_quotes = line[2:-1]
        return stripped_of_quotes

    def _standard_extraction(self, value: str, match_group: str):
        """For most macro metadata values, this extracts the required data"""
        if isinstance(self.parse_results[value], str):
            self.parse_results[value] = match_group
            if value == "comment":
                self._cleanup_comment()
            elif value == "license":
                self._cleanup_license()
        elif isinstance(self.parse_results[value], list):
            self.parse_results[value] = [of.strip() for of in match_group.split(",")]
        else:
            raise SyntaxError(f"Conflicting data type for {value}")

    def _cleanup_comment(self):
        """Remove HTML from the comment line, and truncate it at 512 characters."""

        self.parse_results["comment"] = re.sub(r"<.*?>", "", self.parse_results["comment"])
        if len(self.parse_results["comment"]) > 512:
            self.parse_results["comment"] = self.parse_results["comment"][:511] + "…"

    def _cleanup_license(self):
        if get_license_manager is not None:
            lm = get_license_manager()
            normed_license = lm.normalize(self.parse_results["license"])
            if not normed_license:
                fci.Console.PrintWarning(
                    f"Failed to normalize license {self.parse_results['license']} for macro '{self.name}'\n"
                )
                return
            elif normed_license != self.parse_results["license"]:
                fci.Console.PrintMessage(
                    f"Normalized license {self.parse_results['license']} for macro '{self.name}' to {normed_license}\n"
                )
                self.parse_results["license"] = normed_license

    def _apply_special_handling(self, key: str, line: str):
        # Macro authors are supposed to be providing strings here, but in some
        # cases they are not doing so. If this is the "__version__" tag, try
        # to apply some special handling to accept numbers, and "__date__"
        if key == "__version__":
            self._process_noncompliant_version(line)
            self.remaining_item_map.pop(key)
            return

        raise SyntaxError(f"Failed to process {key} from {line}")

    def _process_noncompliant_version(self, after_equals):
        if is_float(after_equals):
            self.parse_results["version"] = str(after_equals).strip()
        elif "__date__" in after_equals.lower() and self.parse_results["date"]:
            self.parse_results["version"] = self.parse_results["date"]
        else:
            self.parse_results["version"] = "(Unknown)"
            raise SyntaxError(f"Unrecognized version string {after_equals}")

    @staticmethod
    def _detect_illegal_content(line: str):
        """Raise a syntax error if this line contains something we can't handle"""

        lower_line = line.strip().lower()
        if (lower_line.startswith("u'") or lower_line.startswith("'")) and lower_line.endswith("'"):
            return
        if (lower_line.startswith('"') or lower_line.startswith('u"')) and lower_line.endswith('"'):
            return
        if is_float(lower_line):
            return
        if lower_line == "__date__":
            return
        raise SyntaxError(f"Metadata is expected to be a static string, but got {line}")


# Borrowed from Stack Overflow:
# https://stackoverflow.com/questions/736043/checking-if-a-string-can-be-converted-to-float
def is_float(element: Any) -> bool:
    """Determine whether a given item can be converted to a floating-point number"""
    try:
        float(element)
        return True
    except ValueError:
        return False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
Widgets for editing specific types of DocumentObject properties.
Includes a factory method to create the appropriate widget based on type.
"""

import FreeCAD
import FreeCADGui
from PySide import QtGui, QtCore
from typing import Optional


class BasePropertyEditorWidget(QtGui.QWidget):
    """
    Base class for property editor widgets. Includes a factory method
    to create specific subclasses based on property type.
    """

    # Signal emitted when the underlying property value might have changed
    propertyChanged = QtCore.Signal()

    def __init__(self, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None):
        super().__init__(parent)
        self._obj = obj
        self._prop_name = prop_name
        self._layout = QtGui.QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._editor_widget: QtGui.QWidget = None  # The actual input widget (SpinBox, ComboBox)
        self._editor_mode: int = 0  # Default to editable
        self._is_read_only: bool = False
        self._update_editor_mode()
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)

    def attachTo(self, obj: FreeCAD.DocumentObject, prop_name: Optional[str] = None):
        """Attach the editor to a (potentially different) object/property."""
        self._obj = obj
        self._prop_name = prop_name if prop_name else self._prop_name
        self._update_editor_mode()
        self.updateWidget()  # Ensure widget reflects new state

    def _update_editor_mode(self):
        """Fetch and store the current editor mode for the property."""
        if self._obj and self._prop_name:
            self._editor_mode = self._obj.getEditorMode(self._prop_name)
            self._is_read_only = self._editor_mode == 2
            return
        self._editor_mode = 0
        self._is_read_only = False

    def updateWidget(self):
        """Update the editor widget's display from the object property."""
        # Implementation specific to subclasses
        raise NotImplementedError

    def updateProperty(self):
        """Update the object property from the editor widget's value."""
        # Implementation specific to subclasses
        # Should emit propertyChanged signal if value actually changes
        raise NotImplementedError

    @classmethod
    def for_property(
        cls, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None
    ) -> "BasePropertyEditorWidget":
        """
        Factory method to create the appropriate editor widget subclass.
        """
        if not obj or not hasattr(obj, "getPropertyByName"):
            return LabelPropertyEditorWidget(obj, prop_name, parent)

        prop_value = obj.getPropertyByName(prop_name)
        prop_type = obj.getTypeIdOfProperty(prop_name)

        if isinstance(prop_value, FreeCAD.Units.Quantity):
            return QuantityPropertyEditorWidget(obj, prop_name, parent)
        elif isinstance(prop_value, bool):
            return BoolPropertyEditorWidget(obj, prop_name, parent)
        elif isinstance(prop_value, int):
            return IntPropertyEditorWidget(obj, prop_name, parent)
        elif prop_type == "App::PropertyEnumeration":
            return EnumPropertyEditorWidget(obj, prop_name, parent)
        else:
            # Default to a read-only label for other types
            return LabelPropertyEditorWidget(obj, prop_name, parent)


class QuantityPropertyEditorWidget(BasePropertyEditorWidget):
    """Editor widget for Quantity properties."""

    def __init__(self, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None):
        super().__init__(obj, prop_name, parent)
        ui = FreeCADGui.UiLoader()
        self._editor_widget: FreeCADGui.QuantitySpinBox = ui.createWidget("Gui::QuantitySpinBox")
        self._layout.addWidget(self._editor_widget)
        self.updateWidget()  # Set initial value
        # Connect signal after setting initial value to avoid premature update
        self._editor_widget.editingFinished.connect(self.updateProperty)

    def updateWidget(self):
        value: FreeCAD.Units.Quantity = self._obj.getPropertyByName(self._prop_name)
        # Block signals temporarily to prevent feedback loops
        self._editor_widget.blockSignals(True)
        self._editor_widget.setProperty("value", value)
        self._editor_widget.blockSignals(False)
        self._editor_widget.setEnabled(not self._is_read_only)

    def updateProperty(self):
        current_value = self._obj.getPropertyByName(self._prop_name)
        new_value_str: str = self._editor_widget.property("value").UserString
        new_value = FreeCAD.Units.Quantity(new_value_str)
        if new_value_str != current_value:
            self._obj.setPropertyByName(self._prop_name, new_value)
            self.propertyChanged.emit()


class BoolPropertyEditorWidget(BasePropertyEditorWidget):
    """Editor widget for Boolean properties."""

    def __init__(self, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None):
        super().__init__(obj, prop_name, parent)
        self._editor_widget: QtGui.QComboBox = QtGui.QComboBox()
        self._editor_widget.addItems(["False", "True"])
        self._layout.addWidget(self._editor_widget)
        self.updateWidget()
        self._editor_widget.currentIndexChanged.connect(self._on_index_changed)

    def updateWidget(self):
        value: bool = self._obj.getPropertyByName(self._prop_name)
        self._editor_widget.blockSignals(True)
        self._editor_widget.setCurrentIndex(1 if value else 0)
        self._editor_widget.blockSignals(False)
        self._editor_widget.setEnabled(not self._is_read_only)

    def _on_index_changed(self, index: int):
        """Slot connected to currentIndexChanged signal."""
        if self._is_read_only:
            return
        current_value: bool = self._obj.getPropertyByName(self._prop_name)
        new_value: bool = bool(index)
        if new_value != current_value:
            self._obj.setPropertyByName(self._prop_name, new_value)
            self.propertyChanged.emit()

    def updateProperty(self):
        """Update property based on current widget state (for consistency)."""
        self._on_index_changed(self._editor_widget.currentIndex())


class IntPropertyEditorWidget(BasePropertyEditorWidget):
    """Editor widget for Integer properties."""

    def __init__(self, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None):
        super().__init__(obj, prop_name, parent)
        self._editor_widget: QtGui.QSpinBox = QtGui.QSpinBox()
        self._editor_widget.setMinimum(-2147483648)
        self._editor_widget.setMaximum(2147483647)
        self._layout.addWidget(self._editor_widget)
        self.updateWidget()
        self._editor_widget.editingFinished.connect(self.updateProperty)

    def updateWidget(self):
        value = self._obj.getPropertyByName(self._prop_name)
        self._editor_widget.blockSignals(True)
        self._editor_widget.setValue(value or 0)
        self._editor_widget.blockSignals(False)
        self._editor_widget.setEnabled(not self._is_read_only)

    def updateProperty(self):
        current_value: int = self._obj.getPropertyByName(self._prop_name)
        new_value: int = self._editor_widget.value()
        if new_value != current_value:
            self._obj.setPropertyByName(self._prop_name, new_value)
            self.propertyChanged.emit()


class EnumPropertyEditorWidget(BasePropertyEditorWidget):
    """Editor widget for Enumeration properties."""

    def __init__(self, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None):
        super().__init__(obj, prop_name, parent)
        self._editor_widget: QtGui.QComboBox = QtGui.QComboBox()
        self._layout.addWidget(self._editor_widget)
        self._populate_enum()
        self.updateWidget()
        self._editor_widget.currentIndexChanged.connect(self._on_index_changed)

    def _populate_enum(self):
        self._editor_widget.clear()
        enums: list[str] = self._obj.getEnumerationsOfProperty(self._prop_name)
        self._editor_widget.addItems(enums)

    def attachTo(self, obj: FreeCAD.DocumentObject, prop_name: Optional[str] = None):
        """Override attachTo to repopulate enums if object changes."""
        super().attachTo(obj, prop_name)
        self._populate_enum()  # Repopulate in case enums are different

    def updateWidget(self):
        value: str = self._obj.getPropertyByName(self._prop_name)
        self._editor_widget.blockSignals(True)
        index: int = self._editor_widget.findText(value)
        self._editor_widget.setCurrentIndex(index if index >= 0 else 0)
        self._editor_widget.blockSignals(False)
        self._editor_widget.setEnabled(not self._is_read_only)

    def _on_index_changed(self, index: int):
        """Slot connected to currentIndexChanged signal."""
        if self._is_read_only:
            return
        current_value: str = self._obj.getPropertyByName(self._prop_name)
        new_value: str = self._editor_widget.itemText(index)
        if new_value != current_value:
            self._obj.setPropertyByName(self._prop_name, new_value)
            self.propertyChanged.emit()

    def updateProperty(self):
        """Update property based on current widget state (for consistency)."""
        self._on_index_changed(self._editor_widget.currentIndex())


class LabelPropertyEditorWidget(BasePropertyEditorWidget):
    """Read-only label for unsupported or invalid property types."""

    def __init__(self, obj: FreeCAD.DocumentObject, prop_name: str, parent: QtGui.QWidget = None):
        super().__init__(obj, prop_name, parent)
        self._editor_widget: QtGui.QLabel = QtGui.QLabel("N/A")
        self._editor_widget.setTextInteractionFlags(
            QtGui.Qt.TextSelectableByMouse | QtGui.Qt.TextSelectableByKeyboard
        )
        self._layout.addWidget(self._editor_widget)
        self.updateWidget()

    def updateWidget(self):
        text = "N/A"
        text: str = str(self._obj.getPropertyByName(self._prop_name))
        self._editor_widget.setText(text)

    def updateProperty(self):
        # Read-only, no action needed
        pass



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains FreeCAD commands for the BIM workbench"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class BIM_Welcome:
    def GetResources(self):
        return {
            "Pixmap": "BIM_Welcome.svg",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Welcome", "BIM Welcome Screen"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_Welcome", "Shows the BIM workbench welcome screen"),
        }

    def Activated(self):
        self.form = FreeCADGui.PySideUic.loadUi(":ui/dialogWelcome.ui")

        # handle the tutorial links
        self.form.label_4.linkActivated.connect(self.handleLink)
        self.form.label_7.linkActivated.connect(self.handleLink)

        self.form.adjustSize()

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        # show dialog and run setup dialog afterwards if OK was pressed
        result = self.form.exec_()
        if result:
            FreeCADGui.runCommand("BIM_Setup")

        # remove first time flag
        PARAMS.SetBool("FirstTime", False)

    def handleLink(self, link):
        from PySide import QtCore, QtGui

        if hasattr(self, "form"):
            self.form.hide()
            if "BIM_Start_Tutorial" in link:
                FreeCADGui.runCommand("BIM_Tutorial")
            else:
                # print("Opening link:",link)
                url = QtCore.QUrl(link)
                QtGui.QDesktopServices.openUrl(url)


FreeCADGui.addCommand("BIM_Welcome", BIM_Welcome())



#! python
# (c) 2010 Werner Mayer LGPL

"""
An example for a high-level custom feature object to form a full-parametric distance bolt.
"""

__author__ = "Werner Mayer <wmayer@users.sourceforge.net>"

import FreeCAD, Part, math
from FreeCAD import Base


class DistanceBolt:
    def __init__(self, obj):
        """Add the properties: Length, Edges, Radius, Height"""
        obj.addProperty(
            "App::PropertyInteger", "Edges", "Bolt", "Number of edges of the outline", locked=True
        ).Edges = 6
        obj.addProperty(
            "App::PropertyLength",
            "Length",
            "Bolt",
            "Length of the edges of the outline",
            locked=True,
        ).Length = 10.0
        obj.addProperty(
            "App::PropertyLength", "Radius", "Bolt", "Radius of the inner circle", locked=True
        ).Radius = 4.0
        obj.addProperty(
            "App::PropertyLength", "Height", "Bolt", "Height of the extrusion", locked=True
        ).Height = 20.0
        obj.Proxy = self

    def onChanged(self, fp, prop):
        if prop == "Edges" or prop == "Length" or prop == "Radius" or prop == "Height":
            self.execute(fp)

    def execute(self, fp):
        edges = fp.Edges
        if edges < 3:
            edges = 3
        length = fp.Length
        radius = fp.Radius
        height = fp.Height

        m = Base.Matrix()
        m.rotateZ(math.radians(360.0 / edges))

        # create polygon
        polygon = []
        v = Base.Vector(length, 0, 0)
        for i in range(edges):
            polygon.append(v)
            v = m.multiply(v)
        polygon.append(v)
        wire = Part.makePolygon(polygon)

        # create circle
        circ = Part.makeCircle(radius)

        # Create the face with the polygon as outline and the circle as hole
        face = Part.Face([wire, Part.Wire(circ)])

        # Extrude in z to create the final solid
        extrude = face.extrude(Base.Vector(0, 0, height))
        fp.Shape = extrude


def makeDistanceBolt():
    doc = FreeCAD.activeDocument()
    if doc is None:
        doc = FreeCAD.newDocument()
    bolt = doc.addObject("Part::FeaturePython", "Distance_Bolt")
    bolt.Label = "Distance bolt"
    DistanceBolt(bolt)
    bolt.ViewObject.Proxy = 0



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import os
import time

import FreeCAD

from . import ifc_tools
from . import ifc_psets
from . import ifc_materials
from . import ifc_layers
from . import ifc_status
from . import ifc_types

if FreeCAD.GuiUp:
    import FreeCADGui
    import Arch_rc  # needed to load the Arch icons, noqa: F401


PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/NativeIFC")


def open(filename):
    """Opens an IFC file"""

    from PySide import QtCore  # lazy loading

    name = os.path.splitext(os.path.basename(filename))[0]
    FreeCAD.IsOpeningIFC = True
    doc = FreeCAD.newDocument()
    doc.Label = name
    FreeCAD.setActiveDocument(doc.Name)
    insert(filename, doc.Name, singledoc=None)
    del FreeCAD.IsOpeningIFC
    QtCore.QTimer.singleShot(100, unset_modified)
    return doc


def insert(
    filename,
    docname,
    strategy=None,
    shapemode=None,
    switchwb=None,
    silent=False,
    singledoc=False,
):
    """Inserts an IFC document in a FreeCAD document.
    Singledoc defines if the produced result is a locked document or not. The
    strategy is:
    - When opening IFC files, locked/unlocked depends on the preferences (default locked)
    - When inserting IFC files, always unlocked (an IFC doc object is created)"""

    from PySide import QtCore  # lazy loading

    strategy, shapemode, switchwb = get_options(strategy, shapemode, switchwb, silent)
    if strategy is None:
        print("Aborted.")
        return
    stime = time.time()
    try:
        document = FreeCAD.getDocument(docname)
    except NameError:
        document = FreeCAD.newDocument()
    if singledoc is None:
        singledoc = PARAMS.GetBool("SingleDoc", True)
    if singledoc:
        prj_obj = ifc_tools.convert_document(document, filename, shapemode, strategy)
        QtCore.QTimer.singleShot(100, toggle_lock_on)
    else:
        prj_obj = ifc_tools.create_document_object(document, filename, shapemode, strategy)
        QtCore.QTimer.singleShot(100, toggle_lock_off)
    if PARAMS.GetBool("LoadOrphans", True):
        ifc_tools.load_orphans(prj_obj)
    if not silent and PARAMS.GetBool("LoadMaterials", False):
        ifc_materials.load_materials(prj_obj)
    if PARAMS.GetBool("LoadLayers", False):
        ifc_layers.load_layers(prj_obj)
    if PARAMS.GetBool("LoadPsets", False):
        ifc_psets.load_psets(prj_obj)
    if PARAMS.GetBool("LoadTypes", False):
        ifc_types.load_types(prj_obj)
    document.recompute()
    # print a reference to the IFC file on the console
    if FreeCAD.GuiUp and PARAMS.GetBool("IfcFileToConsole", False):
        if isinstance(prj_obj, FreeCAD.DocumentObject):
            pstr = "FreeCAD.getDocument('{}').{}.Proxy.ifcfile"
            pstr = pstr.format(prj_obj.Document.Name, prj_obj.Name)
        else:
            pstr = "FreeCAD.getDocument('{}').Proxy.ifcfile"
            pstr = pstr.format(prj_obj.Name)
        pstr = "ifcfile = " + pstr
        pstr += " # warning: make sure you know what you are doing when using this!"
        FreeCADGui.doCommand(pstr)
    endtime = "%02d:%02d" % (divmod(round(time.time() - stime, 1), 60))
    fsize = round(os.path.getsize(filename) / 1048576, 2)
    print("Imported", os.path.basename(filename), "(", fsize, "Mb ) in", endtime)
    if FreeCAD.GuiUp and switchwb:
        FreeCADGui.activateWorkbench("BIMWorkbench")
    return document


def get_options(strategy=None, shapemode=None, switchwb=None, silent=False):
    """Shows a dialog to get import options

    shapemode: 0 = full shape
               1 = coin only
               2 = no representation
    strategy:  0 = only root object
               1 = only bbuilding structure,
               2 = all children
    """

    psets = PARAMS.GetBool("LoadPsets", False)
    types = PARAMS.GetBool("LoadTypes", False)
    materials = PARAMS.GetBool("LoadMaterials", False)
    layers = PARAMS.GetBool("LoadLayers", False)
    singledoc = PARAMS.GetBool("SingleDoc", False)
    if strategy is None:
        strategy = PARAMS.GetInt("ImportStrategy", 0)
    if shapemode is None:
        shapemode = PARAMS.GetInt("ShapeMode", 0)
    if switchwb is None:
        switchwb = PARAMS.GetBool("SwitchWB", True)
    if silent:
        return strategy, shapemode, switchwb
    ask = PARAMS.GetBool("AskAgain", True)
    if ask and FreeCAD.GuiUp:
        import FreeCADGui

        dlg = FreeCADGui.PySideUic.loadUi(":/ui/dialogImport.ui")
        dlg.checkSwitchWB.hide()  # TODO see what to do with this...
        dlg.comboStrategy.setCurrentIndex(strategy)
        dlg.comboShapeMode.setCurrentIndex(shapemode)
        dlg.checkSwitchWB.setChecked(switchwb)
        dlg.checkAskAgain.setChecked(ask)
        dlg.checkLoadPsets.setChecked(psets)
        dlg.checkLoadTypes.setChecked(types)
        dlg.checkLoadMaterials.setChecked(materials)
        dlg.checkLoadLayers.setChecked(layers)
        dlg.comboSingleDoc.setCurrentIndex(1 - int(singledoc))
        result = dlg.exec_()
        if not result:
            return None, None, None
        strategy = dlg.comboStrategy.currentIndex()
        shapemode = dlg.comboShapeMode.currentIndex()
        switchwb = dlg.checkSwitchWB.isChecked()
        ask = dlg.checkAskAgain.isChecked()
        psets = dlg.checkLoadPsets.isChecked()
        types = dlg.checkLoadTypes.isChecked()
        materials = dlg.checkLoadMaterials.isChecked()
        layers = dlg.checkLoadLayers.isChecked()
        singledoc = dlg.comboSingleDoc.currentIndex()
        PARAMS.SetInt("ImportStrategy", strategy)
        PARAMS.SetInt("ShapeMode", shapemode)
        PARAMS.SetBool("SwitchWB", switchwb)
        PARAMS.SetBool("AskAgain", ask)
        PARAMS.SetBool("LoadPsets", psets)
        PARAMS.SetBool("LoadTypes", types)
        PARAMS.SetBool("LoadMaterials", materials)
        PARAMS.SetBool("LoadLayers", layers)
        PARAMS.SetBool("SingleDoc", bool(1 - singledoc))
    return strategy, shapemode, switchwb


def get_project_type(silent=False):
    """Gets the type of project to make"""

    ask = PARAMS.GetBool("ProjectAskAgain", True)
    ptype = PARAMS.GetBool("ProjectFull", False)
    if silent:
        return ptype
    if ask and FreeCAD.GuiUp:
        import FreeCADGui

        dlg = FreeCADGui.PySideUic.loadUi(":/ui/dialogCreateProject.ui")
        result = dlg.exec_()
        ask = not (dlg.checkBox.isChecked())
        ptype = bool(result)
        PARAMS.SetBool("ProjectAskAgain", ask)
        PARAMS.SetBool("ProjectFull", ptype)
    return ptype


# convenience functions


def toggle_lock_on():
    ifc_status.on_toggle_lock(True, noconvert=True, setchecked=True)


def toggle_lock_off():
    ifc_status.on_toggle_lock(False, noconvert=True, setchecked=True)


def unset_modified():
    try:
        FreeCADGui.ActiveDocument.Modified = False
    except AttributeError:
        pass



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Generator.rotation as generator
import CAMTests.PathTestUtils as PathTestUtils
import numpy as np

Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class TestPathRotationGenerator(PathTestUtils.PathTestBase):
    def test00(self):
        """Test relAngle function"""
        v = FreeCAD.Vector(0.5, 0.5, 0.5)
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.x), 45))
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.y), 45))
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.z), 45))

        v = FreeCAD.Vector(-0.5, 0.5, 0.5)
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.x), 135))
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.y), -45))
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.z), 45))

        v = FreeCAD.Vector(-0.5, -0.5, -0.5)
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.x), -135))
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.y), -135))
        self.assertTrue(np.isclose(generator.relAngle(v, generator.refAxis.z), -135))

    def test10(self):
        """Test Basic Rotation Generator Return"""
        v1 = FreeCAD.Vector(0.0, 0.0, 1.0)
        args = {
            "normalVector": v1,
            "aMin": -360,
            "aMax": 360,
            "cMin": -360,
            "cMax": 360,
            "compound": True,
        }

        result = generator.generate(**args)

        self.assertTrue(type(result) is list)
        self.assertTrue(len(result) == 1)
        self.assertTrue(type(result[0]) is Path.Command)

        command = result[0]
        self.assertTrue(np.isclose(command.Parameters["A"], 0))
        self.assertTrue(np.isclose(command.Parameters["C"], 0))

        args["compound"] = False
        result = generator.generate(**args)
        self.assertTrue(len(result) == 2)

        Path.Log.debug(result)

    def test20(self):
        """Test non-zero rotation"""
        v1 = FreeCAD.Vector(0.5, 0.5, 0.5)
        args = {
            "normalVector": v1,
            "aMin": -360,
            "aMax": 360,
            "cMin": -360,
            "cMax": 360,
            "compound": True,
        }

        result = generator.generate(**args)

        command = result[0]
        Path.Log.debug(command.Parameters)
        self.assertTrue(np.isclose(command.Parameters["A"], 54.736))
        self.assertTrue(np.isclose(command.Parameters["C"], 45))

        Path.Log.track(result)

    def test30(self):
        """Test A limits"""
        v1 = FreeCAD.Vector(0.5, 0.5, 0.5)

        args = {"normalVector": v1, "cMin": -360, "cMax": 360, "compound": True}

        # Constrain a axis rotation negative
        args["aMin"] = -90
        args["aMax"] = 0

        result = generator.generate(**args)
        Path.Log.debug(result)

        command = result[0]
        self.assertTrue(np.isclose(command.Parameters["A"], -54.736))
        self.assertTrue(np.isclose(command.Parameters["C"], -135))

        # Constrain a axis rotation positive
        args["aMin"] = 0
        args["aMax"] = 90

        result = generator.generate(**args)
        Path.Log.debug(result)

        command = result[0]
        self.assertTrue(np.isclose(command.Parameters["A"], 54.736))
        self.assertTrue(np.isclose(command.Parameters["C"], 45))

    def test40(self):
        """Test C limits"""
        v1 = FreeCAD.Vector(0.5, 0.5, 0.5)

        args = {"normalVector": v1, "aMin": -360, "aMax": 360, "compound": True}

        # Constrain a axis rotation negative
        args["cMin"] = -180
        args["cMax"] = 0

        result = generator.generate(**args)
        Path.Log.debug(result)

        command = result[0]
        self.assertTrue(np.isclose(command.Parameters["A"], -54.736))
        self.assertTrue(np.isclose(command.Parameters["C"], -135))

        # Constrain a axis rotation positive
        args["cMin"] = 0
        args["cMax"] = 180

        result = generator.generate(**args)
        Path.Log.debug(result)

        command = result[0]
        self.assertTrue(np.isclose(command.Parameters["A"], 54.736))
        self.assertTrue(np.isclose(command.Parameters["C"], 45))

    def test50(self):
        """Test handling of no valid solution"""
        v1 = FreeCAD.Vector(0.5, 0.5, 0.5)
        args = {
            "normalVector": v1,
            "aMin": 0,
            "aMax": 10,
            "cMin": 0,
            "cMax": 10,
            "compound": True,
        }

        self.assertRaises(ValueError, generator.generate, **args)



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mystran add force constraint"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{


def add_con_force(f, model, mystran_writer):

    # generate pyNastran code
    # force card
    scale_factors = []
    load_ids = []
    force_code = "# force cards, mesh node loads\n"
    for i, femobj in enumerate(mystran_writer.member.cons_force):

        sid = i + 2  # 1 will be the id of the load card
        scale_factors.append(1.0)
        load_ids.append(sid)
        force_obj = femobj["Object"]
        # print(force_obj.Name)

        force_code += f"# {force_obj.Name}\n"
        dirvec = femobj["Object"].DirectionVector
        print(femobj["NodeLoadTable"])
        for ref_shape in femobj["NodeLoadTable"]:
            force_code += f"# {ref_shape[0]}\n"
            for n in sorted(ref_shape[1]):
                # the loads in ref_shape[1][n] are without unit
                node_load = ref_shape[1][n]
                force_code += "model.add_force(sid={}, node={}, mag={}, xyz=({}, {}, {}))\n".format(
                    sid, n, node_load, dirvec.x, dirvec.y, dirvec.z
                )
        force_code += "\n"

    # generate calce factors lists
    # load card, static load combinations
    load_code = "model.add_load(sid=1, scale=1.0, scale_factors={}, load_ids={})\n\n\n".format(
        scale_factors, load_ids
    )

    pynas_code = f"{force_code}\n{load_code}"
    # print(pynas_code)

    # write the pyNastran code
    f.write(pynas_code)

    # execute pyNastran code to add data to the model
    # print(model.get_bdf_stats())
    exec(pynas_code)
    # print(model.get_bdf_stats())

    return model


##  @}



#! python
# (c) 2011 Werner Mayer LGPL

"""
An example for a high-level custom feature object to form a full-parametric parallelepiped.
"""

__author__ = "Werner Mayer <wmayer@users.sourceforge.net>"

import FreeCAD, Part, math
from FreeCAD import Base


class Parallelepiped:
    def __init__(self, obj):
        """Add the properties: Length, Edges, Radius, Height"""
        obj.addProperty("App::PropertyVector", "A", "Parallelepiped", "Vector", locked=True).A = (
            Base.Vector(1, 0, 0)
        )
        obj.addProperty("App::PropertyVector", "B", "Parallelepiped", "Vector", locked=True).B = (
            Base.Vector(0, 1, 0)
        )
        obj.addProperty("App::PropertyVector", "C", "Parallelepiped", "Vector", locked=True).C = (
            Base.Vector(0, 0, 1)
        )
        obj.Proxy = self

    def onChanged(self, fp, prop):
        if prop == "A" or prop == "B" or prop == "C":
            self.execute(fp)

    def execute(self, fp):
        a = fp.A
        b = fp.B
        c = fp.C

        m = Base.Matrix()
        m.A11 = a.x
        m.A12 = a.y
        m.A13 = a.z
        m.A21 = b.x
        m.A22 = b.y
        m.A23 = b.z
        m.A31 = c.x
        m.A32 = c.y
        m.A33 = c.z
        box = Part.makeBox(1, 1, 1)
        fp.Shape = box.transformGeometry(m)


class BoxCylinder:
    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyFloat", "Length", "BoxCylinder", "Length", locked=True
        ).Length = 10.0
        obj.addProperty(
            "App::PropertyFloat", "Width", "BoxCylinder", "Width", locked=True
        ).Width = 10.0
        obj.addProperty(
            "App::PropertyLink", "Source", "BoxCylinder", "Source", locked=True
        ).Source = None
        obj.Proxy = self

    def onChanged(self, fp, prop):
        if prop == "Length" or prop == "Width":
            self.execute(fp)

    def execute(self, fp):
        FreeCAD.Console.PrintMessage(str(fp.Source) + "\n")
        if fp.Source is None:
            return
        r = fp.Source.Radius
        l = fp.Length
        w = fp.Width
        h = 2 * r + 10
        fp.Shape = Part.makeBox(l, w, h)


def makeParallelepiped():
    doc = FreeCAD.activeDocument()
    if doc is None:
        doc = FreeCAD.newDocument()
    obj = doc.addObject("Part::FeaturePython", "Parallelepiped")
    obj.Label = "Parallelepiped"
    Parallelepiped(obj)
    obj.ViewObject.Proxy = 0


def makeBoxCylinder():
    doc = FreeCAD.activeDocument()
    if doc is None:
        doc = FreeCAD.newDocument()
    cyl = doc.addObject("Part::Cylinder", "Cylinder")
    cyl.Radius = 16.0
    cyl.Height = 800.0
    obj = doc.addObject("Part::FeaturePython", "Box")
    BoxCylinder(obj)
    obj.Source = cyl
    obj.Length = 800.0
    obj.Width = 600.0
    obj.ViewObject.Proxy = 0
    doc.recompute()



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

__title__ = "BOPTools.Utils module"
__author__ = "DeepSOIC"
__url__ = "https://www.freecad.org"
__doc__ = "Utility code, used by various modules of BOPTools."


class HashableShape(object):
    "Decorator for Part.Shape, that can be used as key in dicts. Based on isSame method."

    def __init__(self, shape):
        self.Shape = shape
        self.hash = shape.hashCode()

    def __eq__(self, other):
        return self.Shape.isSame(other.Shape)

    def __hash__(self):
        return self.hash


class HashableShape_Deep(object):
    """Similar to HashableShape, except that the things the shape is composed of are compared.

    Example:
        >>> wire2 = Part.Wire(wire1.childShapes())
        >>> wire2.isSame(wire1)
        False # <--- the wire2 is a new wire, although made of edges of wire1
        >>> HashableShape_Deep(wire2) == HashableShape_Deep(wire1)
        True # <--- made of same set of elements
    """

    def __init__(self, shape):
        self.Shape = shape
        self.hash = 0
        for el in shape.childShapes():
            self.hash = self.hash ^ el.hashCode()

    def __eq__(self, other):
        # avoiding extensive comparison for now. Just doing a few extra tests should reduce the already-low chances of false-positives
        if self.hash == other.hash:
            if len(self.Shape.childShapes()) == len(other.Shape.childShapes()):
                if self.Shape.ShapeType == other.Shape.ShapeType:
                    return True
        return False

    def __hash__(self):
        return self.hash


def compoundLeaves(shape_or_compound):
    """compoundLeaves(shape_or_compound): extracts all non-compound shapes from a nested compound.
    Note: shape_or_compound may be a non-compound; then, it is the only thing in the
    returned list."""

    if shape_or_compound.ShapeType == "Compound":
        leaves = []
        for child in shape_or_compound.childShapes():
            leaves.extend(compoundLeaves(child))
        return leaves
    else:
        return [shape_or_compound]


def upgradeToAggregateIfNeeded(list_of_shapes, types=None):
    """upgradeToAggregateIfNeeded(list_of_shapes, types = None): upgrades non-aggregate type
    shapes to aggregate-type shapes if the list has a mix of aggregate and non-aggregate
    type shapes. Returns the new list. Recursively traverses into compounds.

    aggregate shape types are Wire, Shell, CompSolid
    non-aggregate shape types are Vertex, Edge, Face, Solid
    Compounds are something special: they are recursively traversed to upgrade the
    contained shapes.

    Examples:
    list_of_shapes contains only faces -> nothing happens
    list_of_shapes contains faces and shells -> faces are converted to shells

    'types' argument is needed for recursive traversal. Do not supply."""

    import Part

    if types is None:
        types = set()
    for shape in list_of_shapes:
        types.add(shape.ShapeType)
        subshapes = compoundLeaves(shape)
        for subshape in subshapes:
            types.add(subshape.ShapeType)
    if "Wire" in types:
        list_of_shapes = [
            (Part.Wire([shape]) if shape.ShapeType == "Edge" else shape) for shape in list_of_shapes
        ]
    if "Shell" in types:
        list_of_shapes = [
            (Part.makeShell([shape]) if shape.ShapeType == "Face" else shape)
            for shape in list_of_shapes
        ]
    if "CompSolid" in types:
        list_of_shapes = [
            (Part.CompSolid([shape]) if shape.ShapeType == "Solid" else shape)
            for shape in list_of_shapes
        ]
    if "Compound" in types:
        list_of_shapes = [
            (
                Part.makeCompound(upgradeToAggregateIfNeeded(shape.childShapes(), types))
                if shape.ShapeType == "Compound"
                else shape
            )
            for shape in list_of_shapes
        ]
    return list_of_shapes


# adapted from http://stackoverflow.com/a/3603824/6285007
class FrozenClass(object):
    """FrozenClass: prevents adding new attributes to class outside of __init__"""

    __isfrozen = False

    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError(
                "{cls} has no attribute {attr}".format(cls=self.__class__.__name__, attr=key)
            )
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

    def _unfreeze(self):
        self.__isfrozen = False



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mystran add fixed constraint"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{


def add_con_fixed(f, model, mystran_writer):

    # generate pyNastran code
    # spc1 card
    spc_ids = []
    fixed_code = "# spc1 card, Defines a set of single-point constraints\n"
    for i, femobj in enumerate(mystran_writer.member.cons_fixed):

        conid = i + 2  # 1 will be the conid of the spcadd card
        spc_ids.append(conid)
        fixed_obj = femobj["Object"]
        # print(fixed_obj.Name)
        fixed_code += f"# {fixed_obj.Name}\n"
        # node set
        fixed_code += "nodes_{} = {}\n".format(fixed_obj.Name, femobj["Nodes"])
        # all nodes in one line may be to long ... FIXME
        fixed_code += "model.add_spc1(conid={}, components={}, nodes=nodes_{})\n\n".format(
            conid, "123456", fixed_obj.Name
        )

    # spcadd card
    spcadd_code = "# spcadd card, Single-Point Constraint Set Combination from SPC or SPC1 cards\n"
    spcadd_code += f"model.add_spcadd(conid=1, sets={spc_ids})\n\n"

    pynas_code = f"{fixed_code}\n{spcadd_code}"
    # print(pynas_code)

    # write the pyNastran code
    f.write(pynas_code)

    # execute pyNastran code to add data to the model
    # print(model.get_bdf_stats())
    exec(pynas_code)
    # print(model.get_bdf_stats())
    return model


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2022 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Generator.dogboneII as dogboneII
import Path.Base.Language as PathLanguage
import CAMTests.PathTestUtils as PathTestUtils
import math


# Path.Log.setLevel(Path.Log.Level.DEBUG)
Path.Log.setLevel(Path.Log.Level.NOTICE)

PI = math.pi
DebugMode = Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG


def createKinks(maneuver):
    k = []
    moves = maneuver.getMoves()
    if moves:
        move0 = moves[0]
        prev = move0
        for m in moves[1:]:
            k.append(dogboneII.Kink(prev, m))
            prev = m
        if Path.Geom.pointsCoincide(move0.positionBegin(), prev.positionEnd()):
            k.append(dogboneII.Kink(prev, move0))
    return k


def findDogboneKinks(maneuver, threshold):
    if threshold > 0:
        return [k for k in createKinks(maneuver) if k.deflection() > threshold]
    if threshold < 0:
        return [k for k in createKinks(maneuver) if k.deflection() < threshold]
    return createKinks(maneuver)


def MNVR(gcode, begin=None):
    # 'turns out the replace() isn't really necessary
    # leave it here anyway for clarity
    return PathLanguage.Maneuver.FromGCode(gcode.replace("/", "\n"), begin)


def INSTR(gcode, begin=None):
    return MNVR(gcode, begin).instr[0]


def KINK(gcode, begin=None):
    maneuver = MNVR(gcode, begin)
    if len(maneuver.instr) != 2:
        return None
    return dogboneII.Kink(maneuver.instr[0], maneuver.instr[1])


def GEN(generator, length):
    return generator(lambda k, a, n, c: n, length, 1)


class TestGeneratorDogboneII(PathTestUtils.PathTestBase):
    """Unit tests for the dogboneII generator."""

    def assertKinks(self, maneuver, s):
        kinks = [f"{k.deflection():4.2f}" for k in createKinks(maneuver)]
        self.assertEqual(f"[{', '.join(kinks)}]", s)

    def assertBones(self, maneuver, threshold, s):
        bones = [f"({int(b.x())},{int(b.y())})" for b in findDogboneKinks(maneuver, threshold)]
        self.assertEqual(f"[{', '.join(bones)}]", s)

    def assertBone(self, bone, s, digits=0):
        if DebugMode and FreeCAD.GuiUp:
            Path.show(dogboneII.kink_to_path(bone.kink))
            FreeCAD.ActiveDocument.Objects[-1].Visibility = False
            Path.show(dogboneII.bone_to_path(bone))
            FreeCAD.ActiveDocument.Objects[-1].Visibility = False
        Path.Log.debug(f"{bone.kink} : {bone.angle / PI:.2f}")

        b = [i.str(digits) for i in bone.instr]
        self.assertEqual(f"[{', '.join(b)}]", s)

    def test20(self):
        """Verify kinks of maneuvers"""
        self.assertKinks(MNVR("G1X1/G1Y1"), "[1.57]")
        self.assertKinks(MNVR("G1X1/G1Y-1"), "[-1.57]")
        self.assertKinks(MNVR("G1X1/G1Y1/G1X0"), "[1.57, 1.57]")
        self.assertKinks(MNVR("G1X1/G1Y1/G1X0/G1Y0"), "[1.57, 1.57, 1.57, 1.57]")

        self.assertKinks(MNVR("G1Y1/G1X1"), "[-1.57]")
        self.assertKinks(MNVR("G1Y1/G1X1/G1Y0"), "[-1.57, -1.57]")
        self.assertKinks(MNVR("G1Y1/G1X1/G1Y0/G1X0"), "[-1.57, -1.57, -1.57, -1.57]")

        # tangential arc moves
        self.assertKinks(MNVR("G1X1/G3Y2J1"), "[0.00]")
        self.assertKinks(MNVR("G1X1/G3Y2J1G1X0"), "[0.00, 0.00]")

        # folding back arc moves
        self.assertKinks(MNVR("G1X1/G2Y2J1"), "[-3.14]")
        self.assertKinks(MNVR("G1X1/G2Y2J1G1X0"), "[-3.14, 3.14]")

    def test30(self):
        """Verify dogbone detection"""
        self.assertBones(MNVR("G1X1/G1Y1/G1X0/G1Y0"), PI / 4, "[(1,0), (1,1), (0,1), (0,0)]")
        self.assertBones(MNVR("G1X1/G1Y1/G1X0/G1Y0"), -PI / 4, "[]")

        # no bones on flat angle
        self.assertBones(MNVR("G1X1/G1X3Y1/G1X0/G1Y0"), PI / 4, "[(3,1), (0,1), (0,0)]")
        self.assertBones(MNVR("G1X1/G1X3Y1/G1X0/G1Y0"), -PI / 4, "[]")

        # no bones on tangential arc
        self.assertBones(MNVR("G1X1/G3Y2J1/G1X0/G1Y0"), PI / 4, "[(0,2), (0,0)]")
        self.assertBones(MNVR("G1X1/G3Y2J1/G1X0/G1Y0"), -PI / 4, "[]")

        # a bone on perpendicular arc
        self.assertBones(MNVR("G1X1/G3X3I1/G1Y1/G1X0/G1Y0"), PI / 4, "[(3,1), (0,1), (0,0)]")
        self.assertBones(MNVR("G1X1/G3X3I1/G1Y1/G1X0/G1Y0"), -PI / 4, "[(1,0)]")

    def test40(self):
        """Verify horizontal t-bone creation"""
        # Uses test data from test30, if that broke, this can't succeed

        horizontal = GEN(dogboneII.GeneratorTBoneHorizontal, 1)

        # single move right
        maneuver = MNVR("G1X1/G1Y1")
        kinks = findDogboneKinks(maneuver, PI / 4)
        self.assertEqual(len(kinks), 1)
        k = kinks[0]
        p = k.position()
        self.assertEqual(f"({int(p.x)}, {int(p.y)})", "(1, 0)")
        bone = horizontal.generate(k)
        self.assertBone(bone, "[G1{X: 2}, G1{X: 1}]")

        # full loop CCW
        kinks = findDogboneKinks(MNVR("G1X1/G1Y1/G1X0/G1Y0"), PI / 4)
        bones = [horizontal.generate(k) for k in kinks]
        self.assertEqual(len(bones), 4)
        self.assertBone(bones[0], "[G1{X: 2}, G1{X: 1}]")
        self.assertBone(bones[1], "[G1{X: 2}, G1{X: 1}]")
        self.assertBone(bones[2], "[G1{X: -1}, G1{X: 0}]")
        self.assertBone(bones[3], "[G1{X: -1}, G1{X: 0}]")

        # single move left
        maneuver = MNVR("G1X1/G1Y-1")
        kinks = findDogboneKinks(maneuver, -PI / 4)
        self.assertEqual(len(kinks), 1)
        k = kinks[0]
        p = k.position()
        self.assertEqual(f"({int(p.x)}, {int(p.y)})", "(1, 0)")
        bone = horizontal.generate(k)
        self.assertBone(bone, "[G1{X: 2}, G1{X: 1}]")

        # full loop CW
        kinks = findDogboneKinks(MNVR("G1X1/G1Y-1/G1X0/G1Y0"), -PI / 4)
        bones = [horizontal.generate(k) for k in kinks]
        self.assertEqual(len(bones), 4)
        self.assertBone(bones[0], "[G1{X: 2}, G1{X: 1}]")
        self.assertBone(bones[1], "[G1{X: 2}, G1{X: 1}]")
        self.assertBone(bones[2], "[G1{X: -1}, G1{X: 0}]")
        self.assertBone(bones[3], "[G1{X: -1}, G1{X: 0}]")

        # bones on arcs
        kinks = findDogboneKinks(MNVR("G1X1/G3X3I1/G1Y1/G1X0/G1Y0"), PI / 4)
        bones = [horizontal.generate(k) for k in kinks]
        self.assertEqual(len(bones), 3)
        self.assertBone(bones[0], "[G1{X: 4}, G1{X: 3}]")
        self.assertBone(bones[1], "[G1{X: -1}, G1{X: 0}]")
        self.assertBone(bones[2], "[G1{X: -1}, G1{X: 0}]")

        # bones on arcs
        kinks = findDogboneKinks(MNVR("G1X1/G3X3I1/G1Y1/G1X0/G1Y0"), -PI / 4)
        bones = [horizontal.generate(k) for k in kinks]
        self.assertEqual(len(bones), 1)
        self.assertBone(bones[0], "[G1{X: 2}, G1{X: 1}]")

    def test50(self):
        """Verify vertical t-bone creation"""
        # Uses test data from test30, if that broke, this can't succeed

        vertical = GEN(dogboneII.GeneratorTBoneVertical, 1)

        # single move right
        maneuver = MNVR("G1X1/G1Y1")
        kinks = findDogboneKinks(maneuver, PI / 4)
        self.assertEqual(len(kinks), 1)
        k = kinks[0]
        p = k.position()
        self.assertEqual(f"({int(p.x)}, {int(p.y)})", "(1, 0)")
        bone = vertical.generate(k)
        self.assertBone(bone, "[G1{Y: -1}, G1{Y: 0}]")

        # full loop CCW
        kinks = findDogboneKinks(MNVR("G1X1/G1Y1/G1X0/G1Y0"), PI / 4)
        bones = [vertical.generate(k) for k in kinks]
        self.assertEqual(len(bones), 4)
        self.assertBone(bones[0], "[G1{Y: -1}, G1{Y: 0}]")
        self.assertBone(bones[1], "[G1{Y: 2}, G1{Y: 1}]")
        self.assertBone(bones[2], "[G1{Y: 2}, G1{Y: 1}]")
        self.assertBone(bones[3], "[G1{Y: -1}, G1{Y: 0}]")

        # single move left
        maneuver = MNVR("G1X1/G1Y-1")
        kinks = findDogboneKinks(maneuver, -PI / 4)
        self.assertEqual(len(kinks), 1)
        k = kinks[0]
        p = k.position()
        self.assertEqual(f"({int(p.x)}, {int(p.y)})", "(1, 0)")
        bone = vertical.generate(k)
        self.assertBone(bone, "[G1{Y: 1}, G1{Y: 0}]")

        # full loop CW
        kinks = findDogboneKinks(MNVR("G1X1/G1Y-1/G1X0/G1Y0"), -PI / 4)
        bones = [vertical.generate(k) for k in kinks]
        self.assertEqual(len(bones), 4)
        self.assertBone(bones[0], "[G1{Y: 1}, G1{Y: 0}]")
        self.assertBone(bones[1], "[G1{Y: -2}, G1{Y: -1}]")
        self.assertBone(bones[2], "[G1{Y: -2}, G1{Y: -1}]")
        self.assertBone(bones[3], "[G1{Y: 1}, G1{Y: 0}]")

        # bones on arcs
        kinks = findDogboneKinks(MNVR("G1X1/G3X3I1/G1Y1/G1X0/G1Y0"), PI / 4)
        bones = [vertical.generate(k) for k in kinks]
        self.assertEqual(len(bones), 3)
        self.assertBone(bones[0], "[G1{Y: 2}, G1{Y: 1}]")
        self.assertBone(bones[1], "[G1{Y: 2}, G1{Y: 1}]")
        self.assertBone(bones[2], "[G1{Y: -1}, G1{Y: 0}]")

        # bones on arcs
        kinks = findDogboneKinks(MNVR("G1X1/G3X3I1/G1Y1/G1X0/G1Y0"), -PI / 4)
        bones = [vertical.generate(k) for k in kinks]
        self.assertEqual(len(bones), 1)
        self.assertBone(bones[0], "[G1{Y: 1}, G1{Y: 0}]")

    def test60(self):
        """Verify t-bones on edges"""

        on_short_1 = GEN(dogboneII.GeneratorTBoneOnShort, 1)
        on_short_5 = GEN(dogboneII.GeneratorTBoneOnShort, 5)

        # horizontal short edge
        bone = on_short_1.generate(KINK("G1X1/G1Y2"))
        self.assertBone(bone, "[G1{Y: -1}, G1{Y: 0}]")

        bone = on_short_1.generate(KINK("G1X-1/G1Y2"))
        self.assertBone(bone, "[G1{Y: -1}, G1{Y: 0}]")

        # vertical short edge
        bone = on_short_1.generate(KINK("G1Y1/G1X2"))
        self.assertBone(bone, "[G1{X: -1}, G1{X: 0}]")

        bone = on_short_1.generate(KINK("G1Y1/G1X-2"))
        self.assertBone(bone, "[G1{X: 1}, G1{X: 0}]")

        # some other angle
        bone = on_short_5.generate(KINK("G1X1Y1/G1Y-1"))
        self.assertBone(bone, "[G1{X: -2.5, Y: 4.5}, G1{X: 1.0, Y: 1.0}]", 2)

        bone = on_short_5.generate(KINK("G1X-1Y-1/G1Y1"))
        self.assertBone(bone, "[G1{X: 2.5, Y: -4.5}, G1{X: -1.0, Y: -1.0}]", 2)

        # some other angle
        bone = on_short_5.generate(KINK("G1X2Y1/G1Y-3"))
        self.assertBone(bone, "[G1{X: -0.24, Y: 5.5}, G1{X: 2.0, Y: 1.0}]", 2)

        bone = on_short_5.generate(KINK("G1X-2Y-1/G1Y3"))
        self.assertBone(bone, "[G1{X: 0.24, Y: -5.5}, G1{X: -2.0, Y: -1.0}]", 2)

        # short edge - the 2nd
        bone = on_short_1.generate(KINK("G1Y2/G1X1"))
        self.assertBone(bone, "[G1{Y: 3}, G1{Y: 2}]")
        bone = on_short_1.generate(KINK("G1Y2/G1X-1"))
        self.assertBone(bone, "[G1{Y: 3}, G1{Y: 2}]")

        bone = on_short_5.generate(KINK("G1Y-3/G1X2Y-2"))
        self.assertBone(bone, "[G1{X: 2.2, Y: -7.5}, G1{X: 0.0, Y: -3.0}]", 2)

        bone = on_short_5.generate(KINK("G1Y3/G1X-2Y2"))
        self.assertBone(bone, "[G1{X: -2.2, Y: 7.5}, G1{X: 0.0, Y: 3.0}]", 2)

        # long edge
        on_long_1 = GEN(dogboneII.GeneratorTBoneOnLong, 1)
        on_long_5 = GEN(dogboneII.GeneratorTBoneOnLong, 5)

        bone = on_long_1.generate(
            KINK("G1X2/G1Y1"),
        )
        self.assertBone(bone, "[G1{Y: -1}, G1{Y: 0}]")
        bone = on_long_1.generate(KINK("G1X-2/G1Y1"))
        self.assertBone(bone, "[G1{Y: -1}, G1{Y: 0}]")

        bone = on_long_5.generate(KINK("G1Y-1/G1X2Y0"))
        self.assertBone(bone, "[G1{X: 2.2, Y: -5.5}, G1{X: 0.0, Y: -1.0}]", 2)

        bone = on_long_5.generate(KINK("G1Y1/G1X-2Y0"))
        self.assertBone(bone, "[G1{X: -2.2, Y: 5.5}, G1{X: 0.0, Y: 1.0}]", 2)

    def test70(self):
        """Verify dogbone angles"""
        self.assertRoughly(180 * KINK("G1X1/G1Y+1").normAngle() / PI, -45)
        self.assertRoughly(180 * KINK("G1X1/G1Y-1").normAngle() / PI, 45)

        self.assertRoughly(180 * KINK("G1X1/G1X2Y1").normAngle() / PI, -67.5)
        self.assertRoughly(180 * KINK("G1X1/G1X2Y-1").normAngle() / PI, 67.5)

        self.assertRoughly(180 * KINK("G1Y1/G1X+1").normAngle() / PI, 135)
        self.assertRoughly(180 * KINK("G1Y1/G1X-1").normAngle() / PI, 45)

        self.assertRoughly(180 * KINK("G1X-1/G1Y+1").normAngle() / PI, -135)
        self.assertRoughly(180 * KINK("G1X-1/G1Y-1").normAngle() / PI, 135)

        self.assertRoughly(180 * KINK("G1Y-1/G1X-1").normAngle() / PI, -45)
        self.assertRoughly(180 * KINK("G1Y-1/G1X+1").normAngle() / PI, -135)

    def test71(self):
        """Verify dogbones"""

        dogbone = GEN(dogboneII.GeneratorDogbone, 1)

        bone = dogbone.generate(KINK("G1X1/G1Y1"))
        self.assertBone(bone, "[G1{X: 1.7, Y: -0.71}, G1{X: 1.0, Y: 0.0}]", 2)

        bone = dogbone.generate(KINK("G1X1/G1X3Y-1"))
        self.assertBone(bone, "[G1{X: 1.2, Y: 0.97}, G1{X: 1.0, Y: 0.0}]", 2)

        bone = dogbone.generate(KINK("G1X1Y1/G1X2"))
        self.assertBone(bone, "[G1{X: 0.62, Y: 1.9}, G1{X: 1.0, Y: 1.0}]", 2)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import FreeCADGui
import Path
import Path.Dressup.Boundary as PathDressupPathBoundary
import Path.Dressup.Utils as PathDressup
import PathGui

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


class TaskPanel(object):
    def __init__(self, obj, viewProvider):
        self.obj = obj
        self.viewProvider = viewProvider
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/DressupPathBoundary.ui")
        if obj.Stock:
            self.visibilityBoundary = obj.Stock.ViewObject.Visibility
            obj.Stock.ViewObject.Visibility = True
        else:
            self.visibilityBoundary = False

        self.buttonBox = None
        self.isDirty = False

        self.stockFromBase = None
        self.stockFromExisting = None
        self.stockCreateBox = None
        self.stockCreateCylinder = None
        self.stockEdit = None

    def getStandardButtons(self):
        return (
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Apply | QtGui.QDialogButtonBox.Cancel
        )

    def modifyStandardButtons(self, buttonBox):
        self.buttonBox = buttonBox

    def setDirty(self):
        self.isDirty = True
        self.buttonBox.button(QtGui.QDialogButtonBox.Apply).setEnabled(True)

    def setClean(self):
        self.isDirty = False
        self.buttonBox.button(QtGui.QDialogButtonBox.Apply).setEnabled(False)

    def clicked(self, button):
        # callback for standard buttons
        if button == QtGui.QDialogButtonBox.Apply:
            self.updateDressup()
            FreeCAD.ActiveDocument.recompute()

    def abort(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.cleanup(False)

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.cleanup(True)

    def accept(self):
        if self.isDirty:
            self.updateDressup()
        FreeCAD.ActiveDocument.commitTransaction()
        self.cleanup(True)

    def cleanup(self, gui):
        self.viewProvider.clearTaskPanel()
        if gui:
            FreeCADGui.ActiveDocument.resetEdit()
            FreeCADGui.Control.closeDialog()
            FreeCAD.ActiveDocument.recompute()
            if self.obj.Stock:
                self.obj.Stock.ViewObject.Visibility = self.visibilityBoundary

    def updateDressup(self):
        if self.obj.Inside != self.form.stockInside.isChecked():
            self.obj.Inside = self.form.stockInside.isChecked()
        self.stockEdit.getFields(self.obj)
        self.setClean()

    def updateStockEditor(self, index, force=False):
        import Path.Main.Gui.Job as PathJobGui
        import Path.Main.Stock as PathStock

        def setupFromBaseEdit():
            Path.Log.track(index, force)
            if force or not self.stockFromBase:
                self.stockFromBase = PathJobGui.StockFromBaseBoundBoxEdit(
                    self.obj, self.form, force
                )
            self.stockEdit = self.stockFromBase

        def setupCreateBoxEdit():
            Path.Log.track(index, force)
            if force or not self.stockCreateBox:
                self.stockCreateBox = PathJobGui.StockCreateBoxEdit(self.obj, self.form, force)
            self.stockEdit = self.stockCreateBox

        def setupCreateCylinderEdit():
            Path.Log.track(index, force)
            if force or not self.stockCreateCylinder:
                self.stockCreateCylinder = PathJobGui.StockCreateCylinderEdit(
                    self.obj, self.form, force
                )
            self.stockEdit = self.stockCreateCylinder

        def setupFromExisting():
            Path.Log.track(index, force)
            if force or not self.stockFromExisting:
                self.stockFromExisting = PathJobGui.StockFromExistingEdit(
                    self.obj, self.form, force
                )
            if self.stockFromExisting.candidates(self.obj):
                self.stockEdit = self.stockFromExisting
                return True
            return False

        if index == -1:
            if self.obj.Stock is None or PathJobGui.StockFromBaseBoundBoxEdit.IsStock(self.obj):
                setupFromBaseEdit()
            elif PathJobGui.StockCreateBoxEdit.IsStock(self.obj):
                setupCreateBoxEdit()
            elif PathJobGui.StockCreateCylinderEdit.IsStock(self.obj):
                setupCreateCylinderEdit()
            elif PathJobGui.StockFromExistingEdit.IsStock(self.obj):
                setupFromExisting()
            else:
                Path.Log.error(
                    translate("PathJob", "Unsupported stock object %s") % self.obj.Stock.Label
                )
        else:
            if index == PathJobGui.StockFromBaseBoundBoxEdit.Index:
                setupFromBaseEdit()
            elif index == PathJobGui.StockCreateBoxEdit.Index:
                setupCreateBoxEdit()
            elif index == PathJobGui.StockCreateCylinderEdit.Index:
                setupCreateCylinderEdit()
            elif index == PathJobGui.StockFromExistingEdit.Index:
                if not setupFromExisting():
                    setupFromBaseEdit()
                    index = -1
            else:
                Path.Log.error(
                    translate("PathJob", "Unsupported stock type %s (%d)")
                    % (self.form.stock.currentText(), index)
                )
        self.stockEdit.activate(self.obj, index == -1)

    def setupUi(self):
        self.updateStockEditor(-1, False)
        self.form.stockInside.setChecked(self.obj.Inside)

        self.form.stock.currentIndexChanged.connect(self.updateStockEditor)
        if hasattr(self.form.stockInside, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.stockInside.checkStateChanged.connect(self.setDirty)
        else:  # Qt version < 6.7.0
            self.form.stockInside.stateChanged.connect(self.setDirty)
        self.form.stockExtXneg.textChanged.connect(self.setDirty)
        self.form.stockExtXpos.textChanged.connect(self.setDirty)
        self.form.stockExtYneg.textChanged.connect(self.setDirty)
        self.form.stockExtYpos.textChanged.connect(self.setDirty)
        self.form.stockExtZneg.textChanged.connect(self.setDirty)
        self.form.stockExtZpos.textChanged.connect(self.setDirty)
        self.form.stockBoxLength.textChanged.connect(self.setDirty)
        self.form.stockBoxWidth.textChanged.connect(self.setDirty)
        self.form.stockBoxHeight.textChanged.connect(self.setDirty)
        self.form.stockCylinderRadius.textChanged.connect(self.setDirty)
        self.form.stockCylinderHeight.textChanged.connect(self.setDirty)


class DressupPathBoundaryViewProvider(object):
    def __init__(self, vobj):
        self.attach(vobj)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def attach(self, vobj):
        self.vobj = vobj
        self.obj = vobj.Object
        self.panel = None

    def claimChildren(self):
        return [self.obj.Base, self.obj.Stock]

    def onDelete(self, vobj, args=None):
        if vobj.Object and vobj.Object.Proxy:
            vobj.Object.Proxy.onDelete(vobj.Object, args)
        return True

    def setEdit(self, vobj, mode=0):
        panel = TaskPanel(vobj.Object, self)
        self.setupTaskPanel(panel)
        return True

    def unsetEdit(self, vobj, mode=0):
        if self.panel:
            self.panel.abort()

    def setupTaskPanel(self, panel):
        self.panel = panel
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()

    def clearTaskPanel(self):
        self.panel = None

    def getIcon(self):
        if getattr(PathDressup.baseOp(self.obj), "Active", True):
            return ":/icons/CAM_Dressup.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


def Create(base, name="DressupPathBoundary"):
    FreeCAD.ActiveDocument.openTransaction("Create a Boundary dressup")
    obj = PathDressupPathBoundary.Create(base, name)
    obj.ViewObject.Proxy = DressupPathBoundaryViewProvider(obj.ViewObject)
    obj.Base.ViewObject.Visibility = False
    obj.Stock.ViewObject.Visibility = False
    FreeCAD.ActiveDocument.commitTransaction()
    obj.ViewObject.Document.setEdit(obj.ViewObject, 0)
    return obj


class CommandPathDressupPathBoundary:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupPathBoundary", "Boundary"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupPathBoundary",
                "Creates a boundary dress-up from a selected toolpath",
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            Path.Log.error(
                translate("CAM_DressupPathBoundary", "Please select one toolpath object") + "\n"
            )
            return
        baseObject = selection[0]

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create Path Boundary Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.Boundary")
        FreeCADGui.doCommand(
            "Path.Dressup.Gui.Boundary.Create(App.ActiveDocument.%s)" % baseObject.Name
        )
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_DressupPathBoundary", CommandPathDressupPathBoundary())

Path.Log.notice("Loading PathDressupPathBoundaryGui... done\n")



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Mystran tasks"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import os
import os.path
import subprocess

import FreeCAD


try:
    import hfcMystranNeuIn

    result_reading = True
except Exception:
    FreeCAD.Console.PrintWarning("Module to read results not found.\n")
    result_reading = False


from . import writer
from .. import run
from .. import settings
from femmesh import meshsetsgetter
from femtools import femutils
from femtools import membertools


_inputFileName = None


class Check(run.Check):

    def run(self):
        self.pushStatus("Checking analysis member...\n")
        self.check_mesh_exists()
        self.check_material_exists()
        self.check_material_single()  # no multiple material
        self.check_geos_beamsection_single()  # no multiple beamsection
        self.check_geos_shellthickness_single()  # no multiple shellsection
        self.check_geos_beamsection_and_shellthickness()  # either beams or shells


class Prepare(run.Prepare):

    def run(self):
        global _inputFileName
        self.pushStatus("Preparing solver input...\n")

        # get mesh set data
        # TODO see calculix tasks get mesh set data
        mesh_obj = membertools.get_mesh_to_solve(self.analysis)[0]  # pre check done already
        meshdatagetter = meshsetsgetter.MeshSetsGetter(
            self.analysis,
            self.solver,
            mesh_obj,
            membertools.AnalysisMember(self.analysis),
        )
        meshdatagetter.get_mesh_sets()

        # write solver input
        w = writer.FemInputWriterMystran(
            self.analysis,
            self.solver,
            mesh_obj,
            meshdatagetter.member,
            self.directory,
        )
        path = w.write_solver_input()
        # report to user if task succeeded
        if path != "":
            self.pushStatus("Writing solver input completed.")
        else:
            self.pushStatus("Writing solver input failed.")
            self.fail()
        _inputFileName = os.path.splitext(os.path.basename(path))[0]


class Solve(run.Solve):

    def run(self):
        self.pushStatus("Executing solver...\n")

        infile = _inputFileName + ".bdf"

        # get solver binary
        self.pushStatus("Get solver binary...\n")
        binary = settings.get_binary("Mystran")
        if binary is None:
            self.fail()  # a print has been made in settings module

        # run solver
        self.pushStatus("Executing solver...\n")
        self._process = subprocess.Popen(
            args=[binary, infile],  # pass empty param fails! [binary, "", infile]
            cwd=self.directory,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.signalAbort.add(self._process.terminate)
        self._process.communicate()
        self.signalAbort.remove(self._process.terminate)

        # for chatching the output see CalculiX or Elmer solver tasks module


class Results(run.Results):

    def run(self):
        prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/General")
        if not prefs.GetBool("KeepResultsOnReRun", False):
            self.purge_results()
        if result_reading is True:
            self.load_results()

    def purge_results(self):
        self.pushStatus("Purge existing results...\n")
        # TODO see calculix result tasks
        for m in membertools.get_member(self.analysis, "Fem::FemResultObject"):
            if femutils.is_of_type(m.Mesh, "Fem::MeshResult"):
                self.analysis.Document.removeObject(m.Mesh.Name)
            self.analysis.Document.removeObject(m.Name)
        self.analysis.Document.recompute()

    def load_results(self):
        self.pushStatus("Import new results...\n")
        neu_result_file = os.path.join(self.directory, _inputFileName + ".NEU")
        if os.path.isfile(neu_result_file):
            hfcMystranNeuIn.import_neu(neu_result_file)
            # Workaround to move result object into analysis
            for o in self.analysis.Document.Objects:
                if o.Name == "Displacement0":
                    self.analysis.addObject(o)
                    break
        else:
            # TODO: use solver framework status message system
            FreeCAD.Console.PrintError(f"FEM: No results found at {neu_result_file}!\n")
            self.fail()


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

## @package importOBJ
#  \ingroup ARCH
#  \brief OBJ file format exporter
#
#  This module provides tools to import & export OBJ files.
#  It is an alternative to the standard Mesh OBJ exporter
#  and supports exporting faces with more than 3 vertices
#  and supports object colors / materials

import codecs
import ntpath
import os
from builtins import open as pyopen

import FreeCAD
import Arch
import Draft
import DraftGeomUtils
import Mesh
import MeshPart
import Part

from draftutils import params

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:
    # \cond
    def translate(context, text):
        return text

    # \endcond


def findVert(aVertex, aList):
    "finds aVertex in aList, returns index"
    p = Draft.precision()
    for i in range(len(aList)):
        if round(aVertex.X, p) == round(aList[i].X, p):
            if round(aVertex.Y, p) == round(aList[i].Y, p):
                if round(aVertex.Z, p) == round(aList[i].Z, p):
                    return i
    return None


def getIndices(obj, shape, offsetv, offsetvn):
    "returns a list with 2 lists: vertices and face indexes, offset with the given amount"
    p = Draft.precision()
    vlist = []
    vnlist = []
    elist = []
    flist = []
    hascurve = False
    mesh = None

    if isinstance(shape, Part.Shape):
        for e in shape.Edges:
            try:
                if not isinstance(e.Curve, Part.Line):
                    hascurve = True
            except Exception:  # unimplemented curve type
                hascurve = True
            if hascurve:
                tol = params.get_param("MaxDeviationExport", path="Mod/Mesh")
                mesh = Mesh.Mesh()
                mesh.addFacets(shape.getFaces(tol))
                FreeCAD.Console.PrintWarning(
                    translate("Arch", "Found a shape containing curves, triangulating") + "\n"
                )
                break
    elif isinstance(shape, Mesh.Mesh):
        mesh = shape

    if mesh:
        for v in mesh.Topology[0]:
            vlist.append(
                " " + str(round(v[0], p)) + " " + str(round(v[1], p)) + " " + str(round(v[2], p))
            )

        for vn in mesh.Facets:
            vnlist.append(
                " "
                + str(round(vn.Normal[0], p))
                + " "
                + str(round(vn.Normal[1], p))
                + " "
                + str(round(vn.Normal[2], p))
            )

        for i, vn in enumerate(mesh.Topology[1]):
            flist.append(
                " "
                + str(vn[0] + offsetv)
                + "//"
                + str(i + offsetvn)
                + " "
                + str(vn[1] + offsetv)
                + "//"
                + str(i + offsetvn)
                + " "
                + str(vn[2] + offsetv)
                + "//"
                + str(i + offsetvn)
                + " "
            )
    else:
        for v in shape.Vertexes:
            vlist.append(
                " " + str(round(v.X, p)) + " " + str(round(v.Y, p)) + " " + str(round(v.Z, p))
            )
        if not shape.Faces:
            for e in shape.Edges:
                if DraftGeomUtils.geomType(e) == "Line":
                    ei = " " + str(findVert(e.Vertexes[0], shape.Vertexes) + offsetv)
                    ei += " " + str(findVert(e.Vertexes[-1], shape.Vertexes) + offsetv)
                    elist.append(ei)
        for f in shape.Faces:
            if len(f.Wires) > 1:
                # if we have holes, we triangulate
                tris = f.tessellate(1)
                for fdata in tris[1]:
                    fi = ""
                    for vi in fdata:
                        vdata = Part.Vertex(tris[0][vi])
                        fi += " " + str(findVert(vdata, shape.Vertexes) + offsetv)
                    flist.append(fi)
            else:
                fi = ""
                verts = f.OuterWire.OrderedVertexes
                # Avoid flipped normals:
                if f.Orientation == "Reversed":
                    verts.reverse()
                for vert in verts:
                    ind = findVert(vert, shape.Vertexes)
                    if ind is None:
                        return None, None, None, None
                    fi += " " + str(ind + offsetv)
                flist.append(fi)

    return vlist, vnlist, elist, flist


def export(exportList, filename, colors=None):
    """export(exportList,filename,colors=None):
    Called when freecad exports a file. exportList is a list
    of objects, filename is the .obj file to export (a .mtl
    file with same name will also be created together), and
    optionally colors can be a dict containing ["objectName:colorTuple"]
    pairs for use in non-GUI mode."""

    outfile = codecs.open(filename, "wb", encoding="utf8")
    ver = FreeCAD.Version()
    outfile.write("# FreeCAD v" + ver[0] + "." + ver[1] + " build" + ver[2] + " Arch module\n")
    outfile.write("# https://www.freecad.org\n")
    offsetv = 1
    offsetvn = 1
    objectslist = Draft.get_group_contents(exportList, walls=True, addgroups=True)
    objectslist = Arch.pruneIncluded(objectslist, strict=True)
    filenamemtl = filename[:-4] + ".mtl"
    materials = []
    outfile.write("mtllib " + os.path.basename(filenamemtl) + "\n")

    for obj in objectslist:
        if FreeCAD.GuiUp:
            visible = obj.Visibility
        else:
            visible = True
        if not visible:
            continue

        if not (
            obj.isDerivedFrom("Part::Feature")
            or obj.isDerivedFrom("Mesh::Feature")
            or obj.isDerivedFrom("App::Link")
        ):
            continue

        if obj.isDerivedFrom("App::Link"):
            # Get global placement from a Link:
            obj_place = obj.Placement
            parents = obj.Parents
            if parents:
                doc = obj.Document
                parents = [parents[0][0]] + [
                    doc.getObject(name) for name in parents[0][1].split(".")[:-2]
                ]
                parents.reverse()  # child-parent-grandparent order.
                for parent in parents:
                    obj_place = parent.Placement.multiply(obj_place)
        else:
            obj_place = obj.getGlobalPlacement()

        hires = None
        if FreeCAD.GuiUp:
            if obj.ViewObject.DisplayMode == "HiRes":
                # check if high-resolution object is available
                if hasattr(obj, "HiRes") and obj.HiRes:
                    if obj.HiRes.isDerivedFrom("Mesh::Feature"):
                        hires = obj.HiRes.Mesh.copy()
                    else:
                        hires = obj.HiRes.Shape.copy(False, True)
                    hires.Placement = obj_place.multiply(hires.Placement)
                if (
                    not hires
                    and hasattr(obj, "CloneOf")
                    and obj.CloneOf
                    and hasattr(obj.CloneOf, "HiRes")
                    and obj.CloneOf.HiRes
                ):
                    if obj.CloneOf.HiRes.isDerivedFrom("Mesh::Feature"):
                        hires = obj.CloneOf.HiRes.Mesh.copy()
                    else:
                        hires = obj.CloneOf.HiRes.Shape.copy(False, True)
                    hires.Placement = obj_place.multiply(obj.CloneOf.Placement).multiply(
                        hires.Placement
                    )

        if hires:
            vlist, vnlist, elist, flist = getIndices(obj, hires, offsetv, offsetvn)
        elif hasattr(obj, "Shape") and obj.Shape:
            shape = obj.Shape.copy(False, True)
            shape.Placement = obj_place
            vlist, vnlist, elist, flist = getIndices(obj, shape, offsetv, offsetvn)
        elif hasattr(obj, "Mesh") and obj.Mesh:
            mesh = obj.Mesh.copy()
            mesh.Placement = obj_place
            vlist, vnlist, elist, flist = getIndices(obj, mesh, offsetv, offsetvn)

        if vlist is None:
            FreeCAD.Console.PrintError("Unable to export object " + obj.Label + ". Skipping.\n")
        else:
            offsetv += len(vlist)
            offsetvn += len(vnlist)
            outfile.write("o " + obj.Label + "\n")

            # write material
            m = False
            if hasattr(obj, "Material"):
                if obj.Material:
                    if hasattr(obj.Material, "Material"):
                        outfile.write("usemtl " + obj.Material.Name + "\n")
                        materials.append(obj.Material)
                        m = True
            if not m:
                if colors:
                    if obj.Name in colors:
                        color = colors[obj.Name]
                        if color:
                            if isinstance(color[0], tuple):
                                # this is a diffusecolor. For now, use the first color - #TODO: Support per-face colors
                                color = color[0]
                            # print("found color for obj",obj.Name,":",color)
                            mn = Draft.getrgb(color, testbw=False)[1:]
                            outfile.write("usemtl color_" + mn + "\n")
                            materials.append(("color_" + mn, color, 0))
                elif FreeCAD.GuiUp:
                    if hasattr(obj.ViewObject, "ShapeAppearance") and hasattr(
                        obj.ViewObject, "Transparency"
                    ):
                        mn = Draft.getrgb(obj.ViewObject.ShapeColor, testbw=False)[1:]
                        outfile.write("usemtl color_" + mn + "\n")
                        materials.append(
                            ("color_" + mn, obj.ViewObject.ShapeColor, obj.ViewObject.Transparency)
                        )

            # write geometry
            for v in vlist:
                outfile.write("v" + v + "\n")
            for vn in vnlist:
                outfile.write("vn" + vn + "\n")
            for e in elist:
                outfile.write("l" + e + "\n")
            for f in flist:
                outfile.write("f" + f + "\n")

    outfile.close()
    FreeCAD.Console.PrintMessage(translate("Arch", "Successfully written") + " " + filename + "\n")
    if materials:
        outfile = pyopen(filenamemtl, "w")
        outfile.write("# FreeCAD v" + ver[0] + "." + ver[1] + " build" + ver[2] + " Arch module\n")
        outfile.write("# https://www.freecad.org\n")
        kinds = {
            "AmbientColor": "Ka ",
            "DiffuseColor": "Kd ",
            "SpecularColor": "Ks ",
            "EmissiveColor": "Ke ",
            "Transparency": "Tr ",
            "Dissolve": "d ",
        }
        done = []  # store names to avoid duplicates
        for mat in materials:
            if isinstance(mat, tuple):
                if not mat[0] in done:
                    outfile.write("newmtl " + mat[0] + "\n")
                    outfile.write(
                        "Kd " + str(mat[1][0]) + " " + str(mat[1][1]) + " " + str(mat[1][2]) + "\n"
                    )
                    outfile.write("Tr " + str(mat[2] / 100) + "\n")
                    outfile.write("d " + str(1 - mat[2] / 100) + "\n")
                    done.append(mat[0])
            else:
                if not mat.Name in done:
                    outfile.write("newmtl " + mat.Name + "\n")
                    for prop in kinds:
                        if prop in mat.Material:
                            outfile.write(
                                kinds[prop]
                                + mat.Material[prop].strip("()").replace(",", " ")
                                + "\n"
                            )
                    done.append(mat.Name)
        outfile.write("# Material Count: " + str(len(materials)))
        outfile.close()
        FreeCAD.Console.PrintMessage(
            translate("Arch", "Successfully written") + " " + filenamemtl + "\n"
        )


# return entry after given index from an array or None on array end
def peek(index, array):
    if index < len(array) - 1:
        return array[index + 1].strip()
    else:
        return None


def open(filename):
    "called when freecad wants to open a file"
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    return insert(filename, doc.Name)


def insert(filename, docname):

    meshName = ntpath.basename(filename)
    for i in meshName.split():
        if "." in i:
            i = i.split(".")[0]
    meshName = i
    group = None
    "called when freecad wants to import a file"
    try:
        doc = FreeCAD.getDocument(docname)
        group = doc.addObject("App::DocumentObjectGroup", meshName)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc

    with pyopen(filename, "r", encoding="utf8") as infile:
        verts = []
        medges = []
        edges = []
        facets = []
        activeobject = None
        material = None
        colortable = {}
        content_array = []
        for line in infile:
            line = line.strip()
            while line.endswith("\\"):
                next_line = next(infile).strip()
                line = line.rstrip()[:-1] + " " + next_line
            content_array.append(line)
    activeobjectExists = False
    for line in content_array:
        if line[:2] == "o ":
            activeobjectExists = True
    if not activeobjectExists:
        activeobject = meshName
    for index, line in enumerate(content_array):
        if line[:7] == "mtllib ":
            matlib = os.path.join(os.path.dirname(filename), line[7:])
            if os.path.exists(matlib):
                with pyopen(matlib, "r") as matfile:
                    mname = None
                    color = None
                    trans = None
                    for mline in matfile:
                        mline = mline.strip()
                        if mline[:7] == "newmtl ":
                            if mname and color:
                                colortable[mname] = [color, trans]
                            color = None
                            trans = None
                            mname = mline[7:]
                        elif mline[:3] == "Kd ":
                            color = tuple([float(i) for i in mline[3:].split()])
                        elif mline[:2] == "d ":
                            trans = int((1 - float(mline[2:])) * 100)
                    if mname and color:
                        colortable[mname] = [color, trans]
        elif line[:2] == "o ":
            if activeobject:
                makeMesh(doc, group, activeobject, verts, medges, facets, material, colortable)
            material = None
            medges = []
            facets = []
            activeobject = line[2:]
        elif line[:2] == "v ":
            verts.append([float(i) for i in line[2:].split()])
        elif line[:2] == "f ":
            fa = []
            for i in line[2:].split():
                if "/" in i:
                    i = i.split("/")[0]
                fa.append(int(i))
            facets.append(fa)
        elif line[:2] == "l ":
            edge = []
            for i in line[2:].split():
                if "/" in i:
                    i = i.split("/")[0]
                edge.append(int(i))
            edges.append(edge)
            # combine lines into medges
            l = peek(index, content_array)
            if l == None or l[:2] != "l ":
                medges = edges
                edges = []
        elif line[:7] == "usemtl ":
            material = line[7:]
    if activeobject:
        makeMesh(doc, group, activeobject, verts, medges, facets, material, colortable)
    FreeCAD.Console.PrintMessage(translate("Arch", "Successfully imported") + " " + filename + "\n")
    return doc


def makeMesh(doc, group, activeobject, verts, edges, facets, material, colortable):
    mfacets = []
    if facets:
        for facet in facets:
            if len(facet) > 3:
                vecs = [FreeCAD.Vector(*verts[i - 1]) for i in facet]
                vecs.append(vecs[0])
                pol = Part.makePolygon(vecs)
                try:
                    face = Part.Face(pol)
                except Part.OCCError:
                    print("Skipping non-planar polygon:", vecs)
                else:
                    tris = face.tessellate(1)
                    for tri in tris[1]:
                        mfacets.append([tris[0][i] for i in tri])
            else:
                mfacets.append([verts[i - 1] for i in facet])
    if mfacets:
        mobj = doc.addObject("Mesh::Feature", activeobject)
        mobj.Label = activeobject
        mobj.Mesh = Mesh.Mesh(mfacets)
        if FreeCAD.GuiUp and material and material in colortable:
            mobj.ViewObject.ShapeColor = colortable[material][0]
            if colortable[material][1] is not None:
                mobj.ViewObject.Transparency = colortable[material][1]
        if group:
            group.addObjects([mobj])

    # make polylines from edges
    medges = []
    if edges:
        polyline = []
        for edge in edges:
            # single line
            if len(edge) == 2:
                i1 = edge[0]
                i2 = edge[1]
                if i1 == i2:
                    continue
                if len(polyline) > 0:
                    if polyline[-1] == i1:
                        polyline.append(i2)
                    else:
                        medges.append(polyline)
                        polyline = []
                        polyline.append(i1)
                        polyline.append(i2)
                else:
                    polyline.append(i1)
                    polyline.append(i2)
            else:
                medges.append(edge)
        if len(polyline) > 0:
            medges.append(polyline)
    if medges:
        part = doc.addObject("App::Part", activeobject)
        part.Label = activeobject
        features = []
        for strip in medges:
            points = [FreeCAD.Vector(*verts[i - 1]) for i in strip]
            wire = Draft.make_wire(points)
            if FreeCAD.GuiUp and material and material in colortable:
                wire.ViewObject.ShapeColor = colortable[material][0]
                if colortable[material][1] is not None:
                    wire.ViewObject.Transparency = colortable[material][1]
            features.append(wire)
        part.addObjects(features)
        if group:
            group.addObjects([part])

    doc.recompute()
